# flake8: noqa
"""
File copied from cpython test suite:
https://github.com/python/cpython/blob/3.9/Lib/test/string_tests.py

Common tests shared by test_unicode, test_userstring and test_bytes.
"""

# stdlib
import string
import struct
import sys
from test import support

# third party
import pytest

# syft absolute
from syft.lib.python import List
from syft.lib.python import String


class Sequence:
    def __init__(self, seq="wxyz"):
        self.seq = seq

    def __len__(self):
        return len(self.seq)

    def __getitem__(self, i):
        return self.seq[i]


class BadSeq1(Sequence):
    def __init__(self):
        self.seq = [7, "hello", 123]

    def __str__(self):
        return "{0} {1} {2}".format(*self.seq)


class BadSeq2(Sequence):
    def __init__(self):
        self.seq = ["a", "b", "c"]

    def __len__(self):
        return 8


# These tests are for buffers of values (bytes) and not
# specific to character interpretation, used for bytes objects
# and various string implementations

# The type to be tested
# Change in subclasses to change the behaviour of fixtesttype()
type2test = None

# Whether the "contained items" of the container are integers in
# range(0, 256) (i.e. bytes, bytearray) or strings of length 1
# (str)
contains_bytes = False


# check that obj.method(*args) returns result
def checkequal(result, obj, methodname, *args, **kwargs):
    realresult = getattr(obj, methodname)(*args, **kwargs)
    assert result == realresult
    # if the original is returned make sure that
    # this doesn't happen with subclasses


# check that obj.method(*args) raises exc
def checkraises(exc, obj, methodname, *args):
    with pytest.raises(exc) as e_info:
        getattr(obj, methodname)(*args)
    assert str(e_info) != ""


# call obj.method(*args) without any checks
def checkcall(obj, methodname, *args):
    getattr(obj, methodname)(*args)


@pytest.mark.slow
def test_find():
    checkequal(0, String("abcdefghiabc"), "find", String("abc"))
    checkequal(9, String("abcdefghiabc"), "find", String("abc"), 1)
    checkequal(-1, String("abcdefghiabc"), "find", String("def"), 4)

    checkequal(0, String("abc"), "find", String(""), 0)
    checkequal(3, String("abc"), "find", String(""), 3)
    checkequal(-1, String("abc"), "find", String(""), 4)

    # to check the ability to pass None as defaults
    checkequal(2, String("rrarrrrrrrrra"), "find", String("a"))
    checkequal(12, String("rrarrrrrrrrra"), "find", String("a"), 4)
    checkequal(-1, String("rrarrrrrrrrra"), "find", String("a"), 4, 6)
    checkequal(12, String("rrarrrrrrrrra"), "find", String("a"), 4, None)
    checkequal(2, String("rrarrrrrrrrra"), "find", String("a"), None, 6)

    checkraises(TypeError, String("hello"), "find")

    if contains_bytes:
        checkequal(-1, String("hello"), "find", 42)
    else:
        checkraises(TypeError, String("hello"), "find", 42)

    checkequal(0, String(""), "find", String(""))
    checkequal(-1, String(""), "find", String(""), 1, 1)
    checkequal(-1, String(""), "find", String(""), sys.maxsize, 0)

    checkequal(-1, String(""), "find", String("xx"))
    checkequal(-1, String(""), "find", String("xx"), 1, 1)
    checkequal(-1, String(""), "find", String("xx"), sys.maxsize, 0)

    # issue 7458
    checkequal(-1, String("ab"), "find", String("xxx"), sys.maxsize + 1, 0)

    # For a variety of combinations,
    #    verify that str.find() matches __contains__
    #    and that the found substring is really at that location
    charset = [String(""), String("a"), String("b")]
    digits = 4
    base = len(charset)
    teststrings = set()
    for i in range(base ** digits):
        entry = []
        for j in range(digits):
            i, m = divmod(i, base)
            entry.append(charset[m])
        teststrings.add(String("").join(entry))
    for i in teststrings:
        for j in teststrings:
            loc = i.find(j)
            r1 = loc != -1
            r2 = j in i
            assert r1 == r2
            if loc != -1:
                idx = loc + len(j)
                assert i[loc:idx] == j


@pytest.mark.slow
def test_rfind():
    checkequal(9, String("abcdefghiabc"), "rfind", String("abc"))
    checkequal(12, String("abcdefghiabc"), "rfind", "")
    checkequal(0, String("abcdefghiabc"), "rfind", String("abcd"))
    checkequal(-1, String("abcdefghiabc"), "rfind", String("abcz"))

    checkequal(3, String("abc"), "rfind", String(""), 0)
    checkequal(3, String("abc"), "rfind", String(""), 3)
    checkequal(-1, String("abc"), "rfind", String(""), 4)

    # to check the ability to pass None as defaults
    checkequal(12, String("rrarrrrrrrrra"), "rfind", String("a"))
    checkequal(12, String("rrarrrrrrrrra"), "rfind", String("a"), 4)
    checkequal(-1, String("rrarrrrrrrrra"), "rfind", String("a"), 4, 6)
    checkequal(12, String("rrarrrrrrrrra"), "rfind", String("a"), 4, None)
    checkequal(2, String("rrarrrrrrrrra"), "rfind", String("a"), None, 6)

    checkraises(TypeError, String("hello"), "rfind")

    if contains_bytes:
        checkequal(-1, String("hello"), "rfind", 42)
    else:
        checkraises(TypeError, String("hello"), "rfind", 42)

    # For a variety of combinations,
    #    verify that str.rfind() matches __contains__
    #    and that the found substring is really at that location
    charset = [String(""), String("a"), String("b")]
    digits = 3
    base = len(charset)
    teststrings = set()
    for i in range(base ** digits):
        entry = []
        for j in range(digits):
            i, m = divmod(i, base)
            entry.append(charset[m])
        teststrings.add(String("").join(entry))
    for i in teststrings:
        for j in teststrings:
            loc = i.rfind(j)
            r1 = loc != -1
            r2 = j in i
            assert r1 == r2
            if loc != -1:
                assert i[loc : loc + len(j)] == j  # noqa: E203

    # issue 7458
    checkequal(-1, String("ab"), "rfind", String("xxx"), sys.maxsize + 1, 0)

    # issue #15534
    checkequal(0, String("<......\u043c..."), "rfind", "<")


def test_index():
    checkequal(0, String("abcdefghiabc"), "index", String(""))
    checkequal(3, String("abcdefghiabc"), "index", String("def"))
    checkequal(0, String("abcdefghiabc"), "index", String("abc"))
    checkequal(9, String("abcdefghiabc"), "index", "abc", 1)

    checkraises(ValueError, String("abcdefghiabc"), "index", String("hib"))
    checkraises(ValueError, String("abcdefghiab"), "index", String("abc"), 1)
    checkraises(ValueError, String("abcdefghi"), "index", String("ghi"), 8)
    checkraises(ValueError, String("abcdefghi"), "index", String("ghi"), -1)

    # to check the ability to pass None as defaults
    checkequal(2, String("rrarrrrrrrrra"), "index", String("a"))
    checkequal(12, String("rrarrrrrrrrra"), "index", String("a"), 4)
    checkraises(ValueError, String("rrarrrrrrrrra"), "index", String("a"), 4, 6)
    checkequal(12, String("rrarrrrrrrrra"), "index", String("a"), 4, None)
    checkequal(2, String("rrarrrrrrrrra"), "index", String("a"), None, 6)

    checkraises(TypeError, String("hello"), "index")

    checkraises(ValueError, String("hello"), "index", 42)


def test_rindex():
    checkequal(12, String("abcdefghiabc"), "rindex", String(""))
    checkequal(3, String("abcdefghiabc"), "rindex", String("def"))
    checkequal(9, String("abcdefghiabc"), "rindex", String("abc"))
    checkequal(0, String("abcdefghiabc"), "rindex", String("abc"), 0, -1)

    checkraises(ValueError, String("abcdefghiabc"), "rindex", String("hib"))
    checkraises(ValueError, String("defghiabc"), "rindex", String("def"), 1)
    checkraises(ValueError, String("defghiabc"), "rindex", "abc", 0, -1)
    checkraises(ValueError, String("abcdefghi"), "rindex", String("ghi"), 0, 8)
    checkraises(ValueError, String("abcdefghi"), "rindex", String("ghi"), 0, -1)

    # to check the ability to pass None as defaults
    checkequal(12, String("rrarrrrrrrrra"), "rindex", "a")
    checkequal(12, String("rrarrrrrrrrra"), "rindex", "a", 4)
    checkraises(ValueError, String("rrarrrrrrrrra"), "rindex", String("a"), 4, 6)
    checkequal(12, String("rrarrrrrrrrra"), "rindex", String("a"), 4, None)
    checkequal(2, String("rrarrrrrrrrra"), "rindex", String("a"), None, 6)

    checkraises(TypeError, "hello", "rindex")

    if contains_bytes:
        checkraises(ValueError, "hello", "rindex", 42)
    else:
        checkraises(TypeError, String("hello"), "rindex", 42)


def test_lower():
    checkequal(
        String("hello"),
        String("HeLLo"),
        "lower",
    )
    checkequal(String("hello"), String("hello"), "lower")
    checkraises(TypeError, String("hello"), "lower", 42)


def test_upper():
    checkequal(String("HELLO"), String("HeLLo"), "upper")
    checkequal(String("HELLO"), String("HELLO"), "upper")
    checkraises(TypeError, "hello", "upper", 42)


def test_expandtabs():
    checkequal(
        String("abc\rab      def\ng       hi"),
        String("abc\rab\tdef\ng\thi"),
        "expandtabs",
    )
    checkequal(
        String("abc\rab      def\ng       hi"),
        String("abc\rab\tdef\ng\thi"),
        "expandtabs",
        8,
    )
    checkequal("abc\rab  def\ng   hi", String("abc\rab\tdef\ng\thi"), "expandtabs", 4)
    checkequal(
        String("abc\r\nab      def\ng       hi"),
        String("abc\r\nab\tdef\ng\thi"),
        "expandtabs",
    )
    checkequal(
        String("abc\r\nab      def\ng       hi"),
        "abc\r\nab\tdef\ng\thi",
        "expandtabs",
        8,
    )
    checkequal(
        String("abc\r\nab  def\ng   hi"),
        "abc\r\nab\tdef\ng\thi",
        "expandtabs",
        4,
    )
    checkequal(
        String("abc\r\nab\r\ndef\ng\r\nhi"),
        String("abc\r\nab\r\ndef\ng\r\nhi"),
        "expandtabs",
        4,
    )
    # check keyword args
    checkequal(
        String("abc\rab      def\ng       hi"),
        String("abc\rab\tdef\ng\thi"),
        "expandtabs",
        tabsize=8,
    )
    checkequal(
        "abc\rab  def\ng   hi",
        String("abc\rab\tdef\ng\thi"),
        "expandtabs",
        tabsize=4,
    )

    checkequal("  a\n b", String(" \ta\n\tb"), "expandtabs", 1)

    checkraises(TypeError, String("hello"), "expandtabs", 42, 42)
    # This test is only valid when sizeof(int) == sizeof(void*) == 4.
    if sys.maxsize < (1 << 32) and struct.calcsize("P") == 4:
        checkraises(OverflowError, String("\ta\n\tb"), "expandtabs", sys.maxsize)


def test_split():
    # by a char
    checkequal(
        [String("a"), String("b"), "c", String("d")],
        String("a|b|c|d"),
        "split",
        "|",
    )
    checkequal(["a|b|c|d"], String("a|b|c|d"), "split", "|", 0)
    checkequal(
        ["a", String("b|c|d")],
        String("a|b|c|d"),
        "split",
        String("|"),
        1,
    )
    checkequal(["a", "b", "c|d"], String("a|b|c|d"), "split", "|", 2)
    checkequal(["a", "b", "c", "d"], String("a|b|c|d"), "split", String("|"), 3)
    checkequal(
        ["a", String("b"), String("c"), String("d")],
        String("a|b|c|d"),
        "split",
        "|",
        4,
    )
    checkequal(
        ["a", "b", "c", "d"],
        String("a|b|c|d"),
        "split",
        String("|"),
        sys.maxsize - 2,
    )
    checkequal(["a|b|c|d"], String("a|b|c|d"), "split", String("|"), 0)
    checkequal(
        [String("a"), String(""), String("b||c||d")],
        String("a||b||c||d"),
        "split",
        String("|"),
        2,
    )
    checkequal([String("abcd")], String("abcd"), "split", "|")
    checkequal([""], String(""), "split", "|")
    checkequal(
        [String("endcase "), String("")],
        String("endcase |"),
        "split",
        String("|"),
    )
    checkequal(
        ["", String(" startcase")],
        String("| startcase"),
        "split",
        String("|"),
    )
    checkequal(
        [String(""), String("bothcase"), String("")],
        String("|bothcase|"),
        "split",
        "|",
    )
    checkequal(
        ["a", "", String("b\x00c\x00d")],
        String("a\x00\x00b\x00c\x00d"),
        "split",
        "\x00",
        2,
    )

    checkequal(["a"] * 20, (String("a|") * 20)[:-1], "split", String("|"))
    checkequal(
        [String("a")] * 15 + [String("a|a|a|a|a")],
        ("a|" * 20)[:-1],
        "split",
        "|",
        15,
    )

    # by string
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a//b//c//d"),
        "split",
        String("//"),
    )
    checkequal(
        [String("a"), String("b//c//d")],
        String("a//b//c//d"),
        "split",
        "//",
        1,
    )
    checkequal(["a", "b", "c//d"], String("a//b//c//d"), "split", String("//"), 2)
    checkequal(["a", "b", "c", "d"], String("a//b//c//d"), "split", "//", 3)
    checkequal(
        ["a", "b", "c", String("d")],
        String("a//b//c//d"),
        "split",
        "//",
        4,
    )
    checkequal(
        ["a", "b", "c", "d"],
        String("a//b//c//d"),
        "split",
        String("//"),
        sys.maxsize - 10,
    )
    checkequal(
        [String("a//b//c//d")],
        String("a//b//c//d"),
        "split",
        String("//"),
        0,
    )
    checkequal(
        [String("a"), String(""), "b////c////d"],
        String("a////b////c////d"),
        "split",
        "//",
        2,
    )
    checkequal(
        [String(""), String(" bothcase "), String("")],
        String("test bothcase test"),
        "split",
        String("test"),
    )
    checkequal(
        [String("a"), String("bc")],
        String("abbbc"),
        "split",
        String("bb"),
    )
    checkequal(
        [String(""), String("")],
        String("aaa"),
        "split",
        String("aaa"),
    )
    checkequal([String("aaa")], String("aaa"), "split", String("aaa"), 0)
    checkequal(
        [String("ab"), String("ab")],
        String("abbaab"),
        "split",
        String("ba"),
    )
    checkequal([String("aaaa")], String("aaaa"), "split", String("aab"))
    checkequal([String("")], String(""), "split", String("aaa"))
    checkequal([String("aa")], String("aa"), "split", String("aaa"))
    checkequal(
        [String("A"), String("bobb")],
        String("Abbobbbobb"),
        "split",
        String("bbobb"),
    )
    checkequal(
        [String("A"), String("B"), String("")],
        String("AbbobbBbbobb"),
        "split",
        String("bbobb"),
    )

    checkequal(
        [String("a")] * 20,
        (String("aBLAH") * 20)[:-4],
        "split",
        String("BLAH"),
    )
    checkequal(
        [String("a")] * 20,
        (String("aBLAH") * 20)[:-4],
        "split",
        String("BLAH"),
        19,
    )
    checkequal(
        [String("a")] * 18 + [String("aBLAHa")],
        (String("aBLAH") * 20)[:-4],
        "split",
        String("BLAH"),
        18,
    )

    # with keyword args
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a|b|c|d"),
        "split",
        sep=String("|"),
    )
    checkequal(
        [String("a"), String("b|c|d")],
        String("a|b|c|d"),
        "split",
        String("|"),
        maxsplit=1,
    )
    checkequal(
        [String("a"), String("b|c|d")],
        String("a|b|c|d"),
        "split",
        sep=String("|"),
        maxsplit=1,
    )
    checkequal(
        [String("a"), String("b|c|d")],
        String("a|b|c|d"),
        "split",
        maxsplit=1,
        sep=String("|"),
    )
    checkequal(
        [String("a"), String("b c d")],
        String("a b c d"),
        "split",
        maxsplit=1,
    )

    # argument type
    checkraises(TypeError, String("hello"), "split", 42, 42, 42)

    # null case
    checkraises(ValueError, String("hello"), "split", String(""))
    checkraises(ValueError, String("hello"), "split", String(""), 0)


def test_rsplit():
    # by a char
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a|b|c|d"),
        "rsplit",
        String("|"),
    )
    checkequal(
        [String("a|b|c"), String("d")],
        String("a|b|c|d"),
        "rsplit",
        String("|"),
        1,
    )
    checkequal(
        [String("a|b"), String("c"), String("d")],
        String("a|b|c|d"),
        "rsplit",
        String("|"),
        2,
    )
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a|b|c|d"),
        "rsplit",
        String("|"),
        3,
    )
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a|b|c|d"),
        "rsplit",
        String("|"),
        4,
    )
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a|b|c|d"),
        "rsplit",
        String("|"),
        sys.maxsize - 100,
    )
    checkequal(
        [String("a|b|c|d")],
        String("a|b|c|d"),
        "rsplit",
        String("|"),
        0,
    )
    checkequal(
        [String("a||b||c"), String(""), String("d")],
        String("a||b||c||d"),
        "rsplit",
        String("|"),
        2,
    )
    checkequal([String("abcd")], String("abcd"), "rsplit", String("|"))
    checkequal([String("")], String(""), "rsplit", String("|"))
    checkequal(
        [String(""), String(" begincase")],
        String("| begincase"),
        "rsplit",
        String("|"),
    )
    checkequal(
        [String("endcase "), String("")],
        String("endcase |"),
        "rsplit",
        String("|"),
    )
    checkequal(
        [String(""), String("bothcase"), String("")],
        String("|bothcase|"),
        "rsplit",
        String("|"),
    )

    checkequal(
        [String("a\x00\x00b"), String("c"), String("d")],
        String("a\x00\x00b\x00c\x00d"),
        "rsplit",
        String("\x00"),
        2,
    )

    checkequal(
        [String("a")] * 20,
        (String("a|") * 20)[:-1],
        "rsplit",
        String("|"),
    )
    checkequal(
        [String("a|a|a|a|a")] + [String("a")] * 15,
        (String("a|") * 20)[:-1],
        "rsplit",
        String("|"),
        15,
    )

    # by string
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a//b//c//d"),
        "rsplit",
        String("//"),
    )
    checkequal(
        [String("a//b//c"), String("d")],
        String("a//b//c//d"),
        "rsplit",
        String("//"),
        1,
    )
    checkequal(
        [String("a//b"), String("c"), String("d")],
        String("a//b//c//d"),
        "rsplit",
        String("//"),
        2,
    )
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a//b//c//d"),
        "rsplit",
        String("//"),
        3,
    )
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a//b//c//d"),
        "rsplit",
        String("//"),
        4,
    )
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a//b//c//d"),
        "rsplit",
        String("//"),
        sys.maxsize - 5,
    )
    checkequal(
        [String("a//b//c//d")],
        String("a//b//c//d"),
        "rsplit",
        String("//"),
        0,
    )
    checkequal(
        [String("a////b////c"), String(""), String("d")],
        String("a////b////c////d"),
        "rsplit",
        String("//"),
        2,
    )
    checkequal(
        [String(""), String(" begincase")],
        String("test begincase"),
        "rsplit",
        String("test"),
    )
    checkequal(
        [String("endcase "), String("")],
        String("endcase test"),
        "rsplit",
        String("test"),
    )
    checkequal(
        [String(""), String(" bothcase "), String("")],
        String("test bothcase test"),
        "rsplit",
        String("test"),
    )
    checkequal(
        [String("ab"), String("c")],
        String("abbbc"),
        "rsplit",
        String("bb"),
    )
    checkequal(
        [String(""), String("")],
        String("aaa"),
        "rsplit",
        String("aaa"),
    )
    checkequal([String("aaa")], String("aaa"), "rsplit", String("aaa"), 0)
    checkequal(
        [String("ab"), String("ab")],
        String("abbaab"),
        "rsplit",
        String("ba"),
    )
    checkequal([String("aaaa")], String("aaaa"), "rsplit", String("aab"))
    checkequal([String("")], String(""), "rsplit", String("aaa"))
    checkequal([String("aa")], String("aa"), "rsplit", String("aaa"))
    checkequal(
        [String("bbob"), String("A")],
        String("bbobbbobbA"),
        "rsplit",
        String("bbobb"),
    )
    checkequal(
        [String(""), String("B"), String("A")],
        String("bbobbBbbobbA"),
        "rsplit",
        String("bbobb"),
    )

    checkequal(
        [String("a")] * 20,
        (String("aBLAH") * 20)[:-4],
        "rsplit",
        String("BLAH"),
    )
    checkequal(
        [String("a")] * 20,
        (String("aBLAH") * 20)[:-4],
        "rsplit",
        String("BLAH"),
        19,
    )
    checkequal(
        [String("aBLAHa")] + [String("a")] * 18,
        (String("aBLAH") * 20)[:-4],
        "rsplit",
        String("BLAH"),
        18,
    )

    # with keyword args
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a|b|c|d"),
        "rsplit",
        sep=String("|"),
    )
    checkequal(
        [String("a|b|c"), String("d")],
        String("a|b|c|d"),
        "rsplit",
        String("|"),
        maxsplit=1,
    )
    checkequal(
        [String("a|b|c"), String("d")],
        String("a|b|c|d"),
        "rsplit",
        sep=String("|"),
        maxsplit=1,
    )
    checkequal(
        [String("a|b|c"), String("d")],
        String("a|b|c|d"),
        "rsplit",
        maxsplit=1,
        sep=String("|"),
    )
    checkequal(
        [String("a b c"), String("d")],
        String("a b c d"),
        "rsplit",
        maxsplit=1,
    )

    # argument type
    checkraises(TypeError, String("hello"), "rsplit", 42, 42, 42)

    # null case
    checkraises(ValueError, String("hello"), "rsplit", String(""))
    checkraises(ValueError, String("hello"), "rsplit", String(""), 0)


def test_replace():
    EQ = checkequal

    # Operations on the empty string
    EQ(
        String(""),
        String(""),
        "replace",
        String(""),
        String(""),
    )
    EQ(
        String("A"),
        String(""),
        "replace",
        String(""),
        String("A"),
    )
    EQ(
        String(""),
        String(""),
        "replace",
        String("A"),
        String(""),
    )
    EQ(
        String(""),
        String(""),
        "replace",
        String("A"),
        String("A"),
    )
    EQ(
        String(""),
        String(""),
        "replace",
        String(""),
        String(""),
        100,
    )
    EQ(
        String(""),
        String(""),
        "replace",
        String(""),
        String(""),
        sys.maxsize,
    )

    # interleave (from==String(""), 'to' gets inserted everywhere)
    EQ(
        String("A"),
        String("A"),
        "replace",
        String(""),
        String(""),
    )
    EQ(
        String("*A*"),
        String("A"),
        "replace",
        String(""),
        String("*"),
    )
    EQ(
        String("*1A*1"),
        String("A"),
        "replace",
        String(""),
        String("*1"),
    )
    EQ(
        String("*-#A*-#"),
        String("A"),
        "replace",
        String(""),
        String("*-#"),
    )
    EQ(
        String("*-A*-A*-"),
        String("AA"),
        "replace",
        String(""),
        String("*-"),
    )
    EQ(
        String("*-A*-A*-"),
        String("AA"),
        "replace",
        String(""),
        String("*-"),
        -1,
    )
    EQ(
        String("*-A*-A*-"),
        String("AA"),
        "replace",
        String(""),
        String("*-"),
        sys.maxsize,
    )
    EQ(
        String("*-A*-A*-"),
        String("AA"),
        "replace",
        String(""),
        String("*-"),
        4,
    )
    EQ(
        String("*-A*-A*-"),
        String("AA"),
        "replace",
        String(""),
        String("*-"),
        3,
    )
    EQ(
        String("*-A*-A"),
        String("AA"),
        "replace",
        String(""),
        String("*-"),
        2,
    )
    EQ(
        String("*-AA"),
        String("AA"),
        "replace",
        String(""),
        String("*-"),
        1,
    )
    EQ(
        String("AA"),
        String("AA"),
        "replace",
        String(""),
        String("*-"),
        0,
    )

    # single character deletion (from==String("A"), to==String(""))
    EQ(
        String(""),
        String("A"),
        "replace",
        String("A"),
        String(""),
    )
    EQ(
        String(""),
        String("AAA"),
        "replace",
        String("A"),
        String(""),
    )
    EQ(
        String(""),
        String("AAA"),
        "replace",
        String("A"),
        String(""),
        -1,
    )
    EQ(
        String(""),
        String("AAA"),
        "replace",
        String("A"),
        String(""),
        sys.maxsize,
    )
    EQ(
        String(""),
        String("AAA"),
        "replace",
        String("A"),
        String(""),
        4,
    )
    EQ(
        String(""),
        String("AAA"),
        "replace",
        String("A"),
        String(""),
        3,
    )
    EQ(
        String("A"),
        String("AAA"),
        "replace",
        String("A"),
        String(""),
        2,
    )
    EQ(
        String("AA"),
        String("AAA"),
        "replace",
        String("A"),
        String(""),
        1,
    )
    EQ(
        String("AAA"),
        String("AAA"),
        "replace",
        String("A"),
        String(""),
        0,
    )
    EQ(
        String(""),
        String("AAAAAAAAAA"),
        "replace",
        String("A"),
        String(""),
    )
    EQ(
        String("BCD"),
        String("ABACADA"),
        "replace",
        String("A"),
        String(""),
    )
    EQ(
        String("BCD"),
        String("ABACADA"),
        "replace",
        String("A"),
        String(""),
        -1,
    )
    EQ(
        String("BCD"),
        String("ABACADA"),
        "replace",
        String("A"),
        String(""),
        sys.maxsize,
    )
    EQ(
        String("BCD"),
        String("ABACADA"),
        "replace",
        String("A"),
        String(""),
        5,
    )
    EQ(
        String("BCD"),
        String("ABACADA"),
        "replace",
        String("A"),
        String(""),
        4,
    )
    EQ(
        String("BCDA"),
        String("ABACADA"),
        "replace",
        String("A"),
        String(""),
        3,
    )
    EQ(
        String("BCADA"),
        String("ABACADA"),
        "replace",
        String("A"),
        String(""),
        2,
    )
    EQ(
        String("BACADA"),
        String("ABACADA"),
        "replace",
        String("A"),
        String(""),
        1,
    )
    EQ(
        String("ABACADA"),
        String("ABACADA"),
        "replace",
        String("A"),
        String(""),
        0,
    )
    EQ(
        String("BCD"),
        String("ABCAD"),
        "replace",
        String("A"),
        String(""),
    )
    EQ(
        String("BCD"),
        String("ABCADAA"),
        "replace",
        String("A"),
        String(""),
    )
    EQ(
        String("BCD"),
        String("BCD"),
        "replace",
        String("A"),
        String(""),
    )
    EQ(
        String("*************"),
        String("*************"),
        "replace",
        String("A"),
        String(""),
    )
    EQ(
        String("^A^"),
        String("^") + String("A") * 1000 + String("^"),
        "replace",
        String("A"),
        String(""),
        999,
    )

    # substring deletion (from==String("the"), to==String(""))
    EQ(
        String(""),
        String("the"),
        "replace",
        String("the"),
        String(""),
    )
    EQ(
        String("ater"),
        String("theater"),
        "replace",
        String("the"),
        String(""),
    )
    EQ(
        String(""),
        String("thethe"),
        "replace",
        String("the"),
        String(""),
    )
    EQ(
        String(""),
        String("thethethethe"),
        "replace",
        String("the"),
        String(""),
    )
    EQ(
        String("aaaa"),
        String("theatheatheathea"),
        "replace",
        String("the"),
        String(""),
    )
    EQ(
        String("that"),
        String("that"),
        "replace",
        String("the"),
        String(""),
    )
    EQ(
        String("thaet"),
        String("thaet"),
        "replace",
        String("the"),
        String(""),
    )
    EQ(
        String("here and re"),
        String("here and there"),
        "replace",
        String("the"),
        String(""),
    )
    EQ(
        String("here and re and re"),
        String("here and there and there"),
        "replace",
        String("the"),
        String(""),
        sys.maxsize,
    )
    EQ(
        String("here and re and re"),
        String("here and there and there"),
        "replace",
        String("the"),
        String(""),
        -1,
    )
    EQ(
        String("here and re and re"),
        String("here and there and there"),
        "replace",
        String("the"),
        String(""),
        3,
    )
    EQ(
        String("here and re and re"),
        String("here and there and there"),
        "replace",
        String("the"),
        String(""),
        2,
    )
    EQ(
        String("here and re and there"),
        String("here and there and there"),
        "replace",
        String("the"),
        String(""),
        1,
    )
    EQ(
        String("here and there and there"),
        String("here and there and there"),
        "replace",
        String("the"),
        String(""),
        0,
    )
    EQ(
        String("here and re and re"),
        String("here and there and there"),
        "replace",
        String("the"),
        String(""),
    )

    EQ(
        String("abc"),
        String("abc"),
        "replace",
        String("the"),
        String(""),
    )
    EQ(
        String("abcdefg"),
        String("abcdefg"),
        "replace",
        String("the"),
        String(""),
    )

    # substring deletion (from==String("bob"), to==String(""))
    EQ(
        String("bob"),
        String("bbobob"),
        "replace",
        String("bob"),
        String(""),
    )
    EQ(
        String("bobXbob"),
        String("bbobobXbbobob"),
        "replace",
        String("bob"),
        String(""),
    )
    EQ(
        String("aaaaaaa"),
        String("aaaaaaabob"),
        "replace",
        String("bob"),
        String(""),
    )
    EQ(
        String("aaaaaaa"),
        String("aaaaaaa"),
        "replace",
        String("bob"),
        String(""),
    )

    # single character replace in place (len(from)==len(to)==1)
    EQ(
        String("Who goes there?"),
        String("Who goes there?"),
        "replace",
        String("o"),
        String("o"),
    )
    EQ(
        String("WhO gOes there?"),
        String("Who goes there?"),
        "replace",
        String("o"),
        String("O"),
    )
    EQ(
        String("WhO gOes there?"),
        String("Who goes there?"),
        "replace",
        String("o"),
        String("O"),
        sys.maxsize,
    )
    EQ(
        String("WhO gOes there?"),
        String("Who goes there?"),
        "replace",
        String("o"),
        String("O"),
        -1,
    )
    EQ(
        String("WhO gOes there?"),
        String("Who goes there?"),
        "replace",
        String("o"),
        String("O"),
        3,
    )
    EQ(
        String("WhO gOes there?"),
        String("Who goes there?"),
        "replace",
        String("o"),
        String("O"),
        2,
    )
    EQ(
        String("WhO goes there?"),
        String("Who goes there?"),
        "replace",
        String("o"),
        String("O"),
        1,
    )
    EQ(
        String("Who goes there?"),
        String("Who goes there?"),
        "replace",
        String("o"),
        String("O"),
        0,
    )

    EQ(
        String("Who goes there?"),
        String("Who goes there?"),
        "replace",
        String("a"),
        String("q"),
    )
    EQ(
        String("who goes there?"),
        String("Who goes there?"),
        "replace",
        String("W"),
        String("w"),
    )
    EQ(
        String("wwho goes there?ww"),
        String("WWho goes there?WW"),
        "replace",
        String("W"),
        String("w"),
    )
    EQ(
        String("Who goes there!"),
        String("Who goes there?"),
        "replace",
        String("?"),
        String("!"),
    )
    EQ(
        String("Who goes there!!"),
        String("Who goes there??"),
        "replace",
        String("?"),
        String("!"),
    )

    EQ(
        String("Who goes there?"),
        String("Who goes there?"),
        "replace",
        String("."),
        String("!"),
    )

    # substring replace in place (len(from)==len(to) > 1)
    EQ(
        String("Th** ** a t**sue"),
        String("This is a tissue"),
        "replace",
        String("is"),
        String("**"),
    )
    EQ(
        String("Th** ** a t**sue"),
        String("This is a tissue"),
        "replace",
        String("is"),
        String("**"),
        sys.maxsize,
    )
    EQ(
        String("Th** ** a t**sue"),
        String("This is a tissue"),
        "replace",
        String("is"),
        String("**"),
        -1,
    )
    EQ(
        String("Th** ** a t**sue"),
        String("This is a tissue"),
        "replace",
        String("is"),
        String("**"),
        4,
    )
    EQ(
        String("Th** ** a t**sue"),
        String("This is a tissue"),
        "replace",
        String("is"),
        String("**"),
        3,
    )
    EQ(
        String("Th** ** a tissue"),
        String("This is a tissue"),
        "replace",
        String("is"),
        String("**"),
        2,
    )
    EQ(
        String("Th** is a tissue"),
        String("This is a tissue"),
        "replace",
        String("is"),
        String("**"),
        1,
    )
    EQ(
        String("This is a tissue"),
        String("This is a tissue"),
        "replace",
        String("is"),
        String("**"),
        0,
    )
    EQ(
        String("cobob"),
        String("bobob"),
        "replace",
        String("bob"),
        String("cob"),
    )
    EQ(
        String("cobobXcobocob"),
        String("bobobXbobobob"),
        "replace",
        String("bob"),
        String("cob"),
    )
    EQ(
        String("bobob"),
        String("bobob"),
        "replace",
        String("bot"),
        String("bot"),
    )

    # replace single character (len(from)==1, len(to)>1)
    EQ(
        String("ReyKKjaviKK"),
        String("Reykjavik"),
        "replace",
        String("k"),
        String("KK"),
    )
    EQ(
        String("ReyKKjaviKK"),
        String("Reykjavik"),
        "replace",
        String("k"),
        String("KK"),
        -1,
    )
    EQ(
        String("ReyKKjaviKK"),
        String("Reykjavik"),
        "replace",
        String("k"),
        String("KK"),
        sys.maxsize,
    )
    EQ(
        String("ReyKKjaviKK"),
        String("Reykjavik"),
        "replace",
        String("k"),
        String("KK"),
        2,
    )
    EQ(
        String("ReyKKjavik"),
        String("Reykjavik"),
        "replace",
        String("k"),
        String("KK"),
        1,
    )
    EQ(
        String("Reykjavik"),
        String("Reykjavik"),
        "replace",
        String("k"),
        String("KK"),
        0,
    )
    EQ(
        String("A----B----C----"),
        String("A.B.C."),
        "replace",
        String("."),
        String("----"),
    )
    # issue #15534
    EQ(
        String("...\u043c......&lt;"),
        String("...\u043c......<"),
        "replace",
        String("<"),
        String("&lt;"),
    )

    EQ(
        String("Reykjavik"),
        String("Reykjavik"),
        "replace",
        String("q"),
        String("KK"),
    )

    # replace substring (len(from)>1, len(to)!=len(from))
    EQ(
        String("ham, ham, eggs and ham"),
        String("spam, spam, eggs and spam"),
        "replace",
        String("spam"),
        String("ham"),
    )
    EQ(
        String("ham, ham, eggs and ham"),
        String("spam, spam, eggs and spam"),
        "replace",
        String("spam"),
        String("ham"),
        sys.maxsize,
    )
    EQ(
        String("ham, ham, eggs and ham"),
        String("spam, spam, eggs and spam"),
        "replace",
        String("spam"),
        String("ham"),
        -1,
    )
    EQ(
        String("ham, ham, eggs and ham"),
        String("spam, spam, eggs and spam"),
        "replace",
        String("spam"),
        String("ham"),
        4,
    )
    EQ(
        String("ham, ham, eggs and ham"),
        String("spam, spam, eggs and spam"),
        "replace",
        String("spam"),
        String("ham"),
        3,
    )
    EQ(
        String("ham, ham, eggs and spam"),
        String("spam, spam, eggs and spam"),
        "replace",
        String("spam"),
        String("ham"),
        2,
    )
    EQ(
        String("ham, spam, eggs and spam"),
        String("spam, spam, eggs and spam"),
        "replace",
        String("spam"),
        String("ham"),
        1,
    )
    EQ(
        String("spam, spam, eggs and spam"),
        String("spam, spam, eggs and spam"),
        "replace",
        String("spam"),
        String("ham"),
        0,
    )

    EQ(
        String("bobob"),
        String("bobobob"),
        "replace",
        String("bobob"),
        String("bob"),
    )
    EQ(
        String("bobobXbobob"),
        String("bobobobXbobobob"),
        "replace",
        String("bobob"),
        String("bob"),
    )
    EQ(
        String("BOBOBOB"),
        String("BOBOBOB"),
        "replace",
        String("bob"),
        String("bobby"),
    )

    checkequal(
        String("one@two!three!"),
        String("one!two!three!"),
        "replace",
        String("!"),
        String("@"),
        1,
    )
    checkequal(
        String("onetwothree"),
        String("one!two!three!"),
        "replace",
        String("!"),
        String(""),
    )
    checkequal(
        String("one@two@three!"),
        String("one!two!three!"),
        "replace",
        String("!"),
        String("@"),
        2,
    )
    checkequal(
        String("one@two@three@"),
        String("one!two!three!"),
        "replace",
        String("!"),
        String("@"),
        3,
    )
    checkequal(
        String("one@two@three@"),
        String("one!two!three!"),
        "replace",
        String("!"),
        String("@"),
        4,
    )
    checkequal(
        String("one!two!three!"),
        String("one!two!three!"),
        "replace",
        String("!"),
        String("@"),
        0,
    )
    checkequal(
        String("one@two@three@"),
        String("one!two!three!"),
        "replace",
        String("!"),
        String("@"),
    )
    checkequal(
        String("one!two!three!"),
        String("one!two!three!"),
        "replace",
        String("x"),
        String("@"),
    )
    checkequal(
        String("one!two!three!"),
        String("one!two!three!"),
        "replace",
        String("x"),
        String("@"),
        2,
    )
    checkequal(
        String("-a-b-c-"),
        String("abc"),
        "replace",
        String(""),
        String("-"),
    )
    checkequal(
        String("-a-b-c"),
        String("abc"),
        "replace",
        String(""),
        String("-"),
        3,
    )
    checkequal(
        String("abc"),
        String("abc"),
        "replace",
        String(""),
        String("-"),
        0,
    )
    checkequal(
        String(""),
        String(""),
        "replace",
        String(""),
        String(""),
    )
    checkequal(
        String("abc"),
        String("abc"),
        "replace",
        String("ab"),
        String("--"),
        0,
    )
    checkequal(
        String("abc"),
        String("abc"),
        "replace",
        String("xy"),
        String("--"),
    )
    # Next three for SF bug 422088: [OSF1 alpha] string.replace(); died with
    # MemoryError due to empty result (platform malloc issue when requesting
    # 0 bytes).
    checkequal(
        String(""),
        String("123"),
        "replace",
        String("123"),
        String(""),
    )
    checkequal(
        String(""),
        String("123123"),
        "replace",
        String("123"),
        String(""),
    )
    checkequal(
        String("x"),
        String("123x123"),
        "replace",
        String("123"),
        String(""),
    )

    checkraises(TypeError, String("hello"), "replace")
    checkraises(TypeError, String("hello"), "replace", 42)


if sys.version_info >= (3, 9):

    def test_removeprefix():
        checkequal("am", String("spam"), "removeprefix", String("sp"))
        checkequal("spamspam", String("spamspamspam"), "removeprefix", String("spam"))
        checkequal("spam", String("spam"), "removeprefix", String("python"))
        checkequal("spam", String("spam"), "removeprefix", String("spider"))
        checkequal("spam", String("spam"), "removeprefix", String("spam and eggs"))

        checkequal("", String(""), "removeprefix", String(""))
        checkequal("", String(""), "removeprefix", String("abcde"))
        checkequal("abcde", String("abcde"), "removeprefix", String(""))
        checkequal("", String("abcde"), "removeprefix", String("abcde"))

        checkraises(TypeError, String("hello"), "removeprefix")
        checkraises(TypeError, String("hello"), "removeprefix", 42)
        checkraises(TypeError, String("hello"), "removeprefix", 42, "h")
        checkraises(TypeError, String("hello"), "removeprefix", "h", 42)
        checkraises(TypeError, String("hello"), "removeprefix", ("he", "l"))

    def test_removesuffix():
        checkequal("sp", String("spam"), "removesuffix", String("am"))
        checkequal("spamspam", String("spamspamspam"), "removesuffix", String("spam"))
        checkequal("spam", String("spam"), "removesuffix", String("python"))
        checkequal("spam", String("spam"), "removesuffix", String("blam"))
        checkequal("spam", String("spam"), "removesuffix", String("eggs and spam"))

        checkequal("", String(""), "removesuffix", String(""))
        checkequal("", String(""), "removesuffix", String("abcde"))
        checkequal("abcde", String("abcde"), "removesuffix", String(""))
        checkequal("", String("abcde"), "removesuffix", String("abcde"))

        checkraises(TypeError, String("hello"), "removesuffix")
        checkraises(TypeError, String("hello"), "removesuffix", 42)
        checkraises(TypeError, String("hello"), "removesuffix", 42, "h")
        checkraises(TypeError, String("hello"), "removesuffix", "h", 42)
        checkraises(TypeError, String("hello"), "removesuffix", ("lo", "l"))


@pytest.mark.xfail
def test_multiple_arguments():
    checkraises(TypeError, String("hello"), "replace", 42, String("h"))
    checkraises(TypeError, String("hello"), "replace", String("h"), 42)


def test_capitalize():
    checkequal(String(" hello "), String(" hello "), "capitalize")
    checkequal(String("Hello "), String("Hello "), "capitalize")
    checkequal(String("Hello "), String("hello "), "capitalize")
    checkequal(String("Aaaa"), String("aaaa"), "capitalize")
    checkequal(String("Aaaa"), String("AaAa"), "capitalize")

    checkraises(TypeError, String("hello"), "capitalize", 42)


def test_additional_split():
    checkequal(
        [
            String("this"),
            String("is"),
            String("the"),
            "split",
            String("function"),
        ],
        String("this is the split function"),
        "split",
    )

    # by whitespace
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a b c d "),
        "split",
    )
    checkequal(
        [String("a"), String("b c d")],
        String("a b c d"),
        "split",
        None,
        1,
    )
    checkequal(
        [String("a"), String("b"), String("c d")],
        String("a b c d"),
        "split",
        None,
        2,
    )
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a b c d"),
        "split",
        None,
        3,
    )
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a b c d"),
        "split",
        None,
        4,
    )
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a b c d"),
        "split",
        None,
        sys.maxsize - 1,
    )
    checkequal([String("a b c d")], String("a b c d"), "split", None, 0)
    checkequal([String("a b c d")], String("  a b c d"), "split", None, 0)
    checkequal(
        [String("a"), String("b"), String("c  d")],
        String("a  b  c  d"),
        "split",
        None,
        2,
    )

    checkequal([], String("         "), "split")
    checkequal([String("a")], String("  a    "), "split")
    checkequal([String("a"), String("b")], String("  a    b   "), "split")
    checkequal(
        [String("a"), String("b   ")],
        String("  a    b   "),
        "split",
        None,
        1,
    )
    checkequal(
        [String("a    b   c   ")],
        String("  a    b   c   "),
        "split",
        None,
        0,
    )
    checkequal(
        [String("a"), String("b   c   ")],
        String("  a    b   c   "),
        "split",
        None,
        1,
    )
    checkequal(
        [String("a"), String("b"), String("c   ")],
        String("  a    b   c   "),
        "split",
        None,
        2,
    )
    checkequal(
        [String("a"), String("b"), String("c")],
        String("  a    b   c   "),
        "split",
        None,
        3,
    )
    checkequal(
        [String("a"), String("b")],
        String("\n\ta \t\r b \v "),
        "split",
    )
    aaa = String(" a ") * 20
    checkequal([String("a")] * 20, aaa, "split")
    checkequal([String("a")] + [aaa[4:]], aaa, "split", None, 1)
    checkequal([String("a")] * 19 + [String("a ")], aaa, "split", None, 19)

    for b in (
        String("arf\tbarf"),
        String("arf\nbarf"),
        String("arf\rbarf"),
        String("arf\fbarf"),
        String("arf\vbarf"),
    ):
        checkequal([String("arf"), String("barf")], b, "split")
        checkequal([String("arf"), String("barf")], b, "split", None)
        checkequal([String("arf"), String("barf")], b, "split", None, 2)


def test_additional_rsplit():
    checkequal(
        [
            String("this"),
            String("is"),
            String("the"),
            "rsplit",
            String("function"),
        ],
        String("this is the rsplit function"),
        "rsplit",
    )

    # by whitespace
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a b c d "),
        "rsplit",
    )
    checkequal(
        [String("a b c"), String("d")],
        String("a b c d"),
        "rsplit",
        None,
        1,
    )
    checkequal(
        [String("a b"), String("c"), String("d")],
        String("a b c d"),
        "rsplit",
        None,
        2,
    )
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a b c d"),
        "rsplit",
        None,
        3,
    )
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a b c d"),
        "rsplit",
        None,
        4,
    )
    checkequal(
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
        String("a b c d"),
        "rsplit",
        None,
        sys.maxsize - 20,
    )
    checkequal([String("a b c d")], String("a b c d"), "rsplit", None, 0)
    checkequal([String("a b c d")], String("a b c d  "), "rsplit", None, 0)
    checkequal(
        [String("a  b"), String("c"), String("d")],
        String("a  b  c  d"),
        "rsplit",
        None,
        2,
    )

    checkequal([], String("         "), "rsplit")
    checkequal([String("a")], String("  a    "), "rsplit")
    checkequal([String("a"), String("b")], String("  a    b   "), "rsplit")
    checkequal(
        [String("  a"), String("b")],
        String("  a    b   "),
        "rsplit",
        None,
        1,
    )
    checkequal(
        [String("  a    b   c")],
        String("  a    b   c   "),
        "rsplit",
        None,
        0,
    )
    checkequal(
        [String("  a    b"), String("c")],
        String("  a    b   c   "),
        "rsplit",
        None,
        1,
    )
    checkequal(
        [String("  a"), String("b"), String("c")],
        String("  a    b   c   "),
        "rsplit",
        None,
        2,
    )
    checkequal(
        [String("a"), String("b"), String("c")],
        String("  a    b   c   "),
        "rsplit",
        None,
        3,
    )
    checkequal(
        [String("a"), String("b")],
        String("\n\ta \t\r b \v "),
        "rsplit",
        None,
        88,
    )
    aaa = String(" a ") * 20
    checkequal([String("a")] * 20, aaa, "rsplit")
    checkequal([aaa[:-4]] + [String("a")], aaa, "rsplit", None, 1)
    checkequal([String(" a  a")] + [String("a")] * 18, aaa, "rsplit", None, 18)

    for b in (
        String("arf\tbarf"),
        String("arf\nbarf"),
        String("arf\rbarf"),
        String("arf\fbarf"),
        String("arf\vbarf"),
    ):
        checkequal([String("arf"), String("barf")], b, "rsplit")
        checkequal([String("arf"), String("barf")], b, "rsplit", None)
        checkequal([String("arf"), String("barf")], b, "rsplit", None, 2)


def test_strip_whitespace():
    checkequal(String("hello"), String("   hello   "), "strip")
    checkequal(String("hello   "), String("   hello   "), "lstrip")
    checkequal(String("   hello"), String("   hello   "), "rstrip")
    checkequal(String("hello"), String("hello"), "strip")

    b = String(" \t\n\r\f\vabc \t\n\r\f\v")
    checkequal(String("abc"), b, "strip")
    checkequal(String("abc \t\n\r\f\v"), b, "lstrip")
    checkequal(String(" \t\n\r\f\vabc"), b, "rstrip")

    # strip/lstrip/rstrip with None arg
    checkequal(String("hello"), String("   hello   "), "strip", None)
    checkequal(String("hello   "), String("   hello   "), "lstrip", None)
    checkequal(String("   hello"), String("   hello   "), "rstrip", None)
    checkequal(String("hello"), String("hello"), "strip", None)


def test_strip():
    # strip/lstrip/rstrip with str arg
    checkequal(
        String("hello"),
        String("xyzzyhelloxyzzy"),
        "strip",
        String("xyz"),
    )
    checkequal(
        String("helloxyzzy"),
        String("xyzzyhelloxyzzy"),
        "lstrip",
        String("xyz"),
    )
    checkequal(
        String("xyzzyhello"),
        String("xyzzyhelloxyzzy"),
        "rstrip",
        String("xyz"),
    )
    checkequal(String("hello"), String("hello"), "strip", String("xyz"))
    checkequal(
        String(""),
        String("mississippi"),
        "strip",
        String("mississippi"),
    )

    # only trim the start and end; does not strip internal characters
    checkequal(
        String("mississipp"),
        String("mississippi"),
        "strip",
        String("i"),
    )

    checkraises(TypeError, String("hello"), "strip", 42, 42)
    checkraises(TypeError, String("hello"), "lstrip", 42, 42)
    checkraises(TypeError, String("hello"), "rstrip", 42, 42)


def test_ljust():
    checkequal(String("abc       "), String("abc"), "ljust", 10)
    checkequal(String("abc   "), String("abc"), "ljust", 6)
    checkequal(String("abc"), String("abc"), "ljust", 3)
    checkequal(String("abc"), String("abc"), "ljust", 2)
    checkequal(
        String("abc*******"),
        String("abc"),
        "ljust",
        10,
        String("*"),
    )
    checkraises(TypeError, String("abc"), "ljust")


def test_rjust():
    checkequal(String("       abc"), String("abc"), "rjust", 10)
    checkequal(String("   abc"), String("abc"), "rjust", 6)
    checkequal(String("abc"), String("abc"), "rjust", 3)
    checkequal(String("abc"), String("abc"), "rjust", 2)
    checkequal(
        String("*******abc"),
        String("abc"),
        "rjust",
        10,
        String("*"),
    )
    checkraises(TypeError, String("abc"), "rjust")


def test_center():
    checkequal(String("   abc    "), String("abc"), "center", 10)
    checkequal(String(" abc  "), String("abc"), "center", 6)
    checkequal(String("abc"), String("abc"), "center", 3)
    checkequal(String("abc"), String("abc"), "center", 2)
    checkequal(
        String("***abc****"),
        String("abc"),
        "center",
        10,
        String("*"),
    )
    checkraises(TypeError, String("abc"), "center")


def test_swapcase():
    checkequal(String("hEllO CoMPuTErS"), String("HeLLo cOmpUteRs"), "swapcase")
    checkraises(TypeError, String("hello"), "swapcase", 42)


def test_zfill():
    checkequal(String("123"), String("123"), "zfill", 2)
    checkequal(String("123"), String("123"), "zfill", 3)
    checkequal(String("0123"), String("123"), "zfill", 4)
    checkequal(String("+123"), String("+123"), "zfill", 3)
    checkequal(String("+123"), String("+123"), "zfill", 4)
    checkequal(String("+0123"), String("+123"), "zfill", 5)
    checkequal(String("-123"), String("-123"), "zfill", 3)
    checkequal(String("-123"), String("-123"), "zfill", 4)
    checkequal(String("-0123"), String("-123"), "zfill", 5)
    checkequal(String("000"), String(""), "zfill", 3)
    checkequal(String("34"), String("34"), "zfill", 1)
    checkequal(String("0034"), String("34"), "zfill", 4)

    checkraises(TypeError, String("123"), "zfill")


def test_islower():
    checkequal(False, String(""), "islower")
    checkequal(True, String("a"), "islower")
    checkequal(False, String("A"), "islower")
    checkequal(False, String("\n"), "islower")
    checkequal(True, String("abc"), "islower")
    checkequal(False, String("aBc"), "islower")
    checkequal(True, String("abc\n"), "islower")
    checkraises(TypeError, String("abc"), "islower", 42)


def test_isupper():
    checkequal(False, String(""), "isupper")
    checkequal(False, String("a"), "isupper")
    checkequal(True, String("A"), "isupper")
    checkequal(False, String("\n"), "isupper")
    checkequal(True, String("ABC"), "isupper")
    checkequal(False, String("AbC"), "isupper")
    checkequal(True, String("ABC\n"), "isupper")
    checkraises(TypeError, String("abc"), "isupper", 42)


def test_istitle():
    checkequal(False, String(""), "istitle")
    checkequal(False, String("a"), "istitle")
    checkequal(True, String("A"), "istitle")
    checkequal(False, String("\n"), "istitle")
    checkequal(True, String("A Titlecased Line"), "istitle")
    checkequal(True, String("A\nTitlecased Line"), "istitle")
    checkequal(True, String("A Titlecased, Line"), "istitle")
    checkequal(False, String("Not a capitalized String"), "istitle")
    checkequal(False, String("Not\ta Titlecase String"), "istitle")
    checkequal(False, String("Not--a Titlecase String"), "istitle")
    checkequal(False, String("NOT"), "istitle")
    checkraises(TypeError, String("abc"), "istitle", 42)


def test_isspace():
    checkequal(False, String(""), "isspace")
    checkequal(False, String("a"), "isspace")
    checkequal(True, String(" "), "isspace")
    checkequal(True, String("\t"), "isspace")
    checkequal(True, String("\r"), "isspace")
    checkequal(True, String("\n"), "isspace")
    checkequal(True, String(" \t\r\n"), "isspace")
    checkequal(False, String(" \t\r\na"), "isspace")
    checkraises(TypeError, String("abc"), "isspace", 42)


def test_isalpha():
    checkequal(False, String(""), "isalpha")
    checkequal(True, String("a"), "isalpha")
    checkequal(True, String("A"), "isalpha")
    checkequal(False, String("\n"), "isalpha")
    checkequal(True, String("abc"), "isalpha")
    checkequal(False, String("aBc123"), "isalpha")
    checkequal(False, String("abc\n"), "isalpha")
    checkraises(TypeError, String("abc"), "isalpha", 42)


def test_isalnum():
    checkequal(False, String(""), "isalnum")
    checkequal(True, String("a"), "isalnum")
    checkequal(True, String("A"), "isalnum")
    checkequal(False, String("\n"), "isalnum")
    checkequal(True, String("123abc456"), "isalnum")
    checkequal(True, String("a1b3c"), "isalnum")
    checkequal(False, String("aBc000 "), "isalnum")
    checkequal(False, String("abc\n"), "isalnum")
    checkraises(TypeError, String("abc"), "isalnum", 42)


def test_isascii():
    if sys.version_info >= (3, 7):
        checkequal(True, String(""), "isascii")
        checkequal(True, String("\x00"), "isascii")
        checkequal(True, String("\x7f"), "isascii")
        checkequal(True, String("\x00\x7f"), "isascii")
        checkequal(False, String("\x80"), "isascii")
        checkequal(False, String("\xe9"), "isascii")
        # bytes.isascii() and bytearray.isascii() has optimization which
        # check 4 or 8 bytes at once.  So check some alignments.
        for p in range(8):
            checkequal(True, String(" ") * p + String("\x7f"), "isascii")
            checkequal(False, String(" ") * p + String("\x80"), "isascii")
            checkequal(
                True,
                String(" ") * p + String("\x7f") + String(" ") * 8,
                "isascii",
            )
            checkequal(
                False,
                String(" ") * p + String("\x80") + String(" ") * 8,
                "isascii",
            )
    else:
        with pytest.raises(AttributeError):
            checkequal(True, String(""), "isascii")


def test_isdigit():
    checkequal(False, String(""), "isdigit")
    checkequal(False, String("a"), "isdigit")
    checkequal(True, String("0"), "isdigit")
    checkequal(True, String("0123456789"), "isdigit")
    checkequal(False, String("0123456789a"), "isdigit")

    checkraises(TypeError, String("abc"), "isdigit", 42)


def test_title():
    checkequal(String(" Hello "), String(" hello "), "title")
    checkequal(String("Hello "), String("hello "), "title")
    checkequal(String("Hello "), String("Hello "), "title")
    checkequal(
        String("Format This As Title String"),
        String("fOrMaT thIs aS titLe String"),
        "title",
    )
    checkequal(
        String("Format,This-As*Title;String"),
        String("fOrMaT,thIs-aS*titLe;String"),
        "title",
    )
    checkequal(String("Getint"), String("getInt"), "title")
    checkraises(TypeError, String("hello"), "title", 42)


def test_splitlines():
    checkequal(
        [
            String("abc"),
            String("def"),
            String(""),
            String("ghi"),
        ],
        String("abc\ndef\n\rghi"),
        "splitlines",
    )
    checkequal(
        [
            String("abc"),
            String("def"),
            String(""),
            String("ghi"),
        ],
        String("abc\ndef\n\r\nghi"),
        "splitlines",
    )
    checkequal(
        [String("abc"), String("def"), String("ghi")],
        String("abc\ndef\r\nghi"),
        "splitlines",
    )
    checkequal(
        [String("abc"), String("def"), String("ghi")],
        String("abc\ndef\r\nghi\n"),
        "splitlines",
    )
    checkequal(
        [
            String("abc"),
            String("def"),
            String("ghi"),
            String(""),
        ],
        String("abc\ndef\r\nghi\n\r"),
        "splitlines",
    )
    checkequal(
        [
            String(""),
            String("abc"),
            String("def"),
            String("ghi"),
            String(""),
        ],
        String("\nabc\ndef\r\nghi\n\r"),
        "splitlines",
    )
    checkequal(
        [
            String(""),
            String("abc"),
            String("def"),
            String("ghi"),
            String(""),
        ],
        String("\nabc\ndef\r\nghi\n\r"),
        "splitlines",
        False,
    )
    checkequal(
        [
            String("\n"),
            String("abc\n"),
            String("def\r\n"),
            String("ghi\n"),
            String("\r"),
        ],
        String("\nabc\ndef\r\nghi\n\r"),
        "splitlines",
        True,
    )
    checkequal(
        [
            String(""),
            String("abc"),
            String("def"),
            String("ghi"),
            String(""),
        ],
        String("\nabc\ndef\r\nghi\n\r"),
        "splitlines",
        keepends=False,
    )
    checkequal(
        [
            String("\n"),
            String("abc\n"),
            String("def\r\n"),
            String("ghi\n"),
            String("\r"),
        ],
        String("\nabc\ndef\r\nghi\n\r"),
        "splitlines",
        keepends=True,
    )

    checkraises(TypeError, String("abc"), "splitlines", 42, 42)


# fixes python <= 3.7
# https://github.com/python/cpython/commit/b015fc86f7b1f35283804bfee788cce0a5495df7
def test_capitalize_nonascii():
    # check that titlecased chars are lowered correctly
    # \u1ffc is the titlecased char
    # \u03a9\u0399
    if sys.version_info >= (3, 8):
        # a, b, capitalize
        # ῼῳῳῳ, ῳῳῼῼ, capitalize
        checkequal(
            String("\u1ffc\u1ff3\u1ff3\u1ff3"),
            String("\u1ff3\u1ff3\u1ffc\u1ffc"),
            "capitalize",
        )
    else:
        # a, b, capitalize
        # ΩΙῳῳῳ, ῳῳῼῼ, capitalize
        checkequal(
            String("\u03a9\u0399\u1ff3\u1ff3\u1ff3"),
            String("\u1ff3\u1ff3\u1ffc\u1ffc"),
            "capitalize",
        )
    # check with cased non-letter chars
    checkequal(
        String("\u24c5\u24e8\u24e3\u24d7\u24de\u24dd"),
        String("\u24c5\u24ce\u24c9\u24bd\u24c4\u24c3"),
        "capitalize",
    )
    checkequal(
        String("\u24c5\u24e8\u24e3\u24d7\u24de\u24dd"),
        String("\u24df\u24e8\u24e3\u24d7\u24de\u24dd"),
        "capitalize",
    )
    checkequal(
        String("\u2160\u2171\u2172"),
        String("\u2160\u2161\u2162"),
        "capitalize",
    )
    checkequal(
        String("\u2160\u2171\u2172"),
        String("\u2170\u2171\u2172"),
        "capitalize",
    )
    # check with Ll chars with no upper - nothing changes here
    checkequal(
        String("\u019b\u1d00\u1d86\u0221\u1fb7"),
        String("\u019b\u1d00\u1d86\u0221\u1fb7"),
        "capitalize",
    )


def test_startswith():
    checkequal(True, String("hello"), "startswith", String("he"))
    checkequal(True, String("hello"), "startswith", String("hello"))
    checkequal(False, String("hello"), "startswith", String("hello world"))
    checkequal(True, String("hello"), "startswith", String(""))
    checkequal(False, String("hello"), "startswith", String("ello"))
    checkequal(True, String("hello"), "startswith", String("ello"), 1)
    checkequal(True, String("hello"), "startswith", String("o"), 4)
    checkequal(False, String("hello"), "startswith", String("o"), 5)
    checkequal(True, String("hello"), "startswith", String(""), 5)
    checkequal(False, String("hello"), "startswith", String("lo"), 6)
    checkequal(True, String("helloworld"), "startswith", String("lowo"), 3)
    checkequal(True, String("helloworld"), "startswith", String("lowo"), 3, 7)
    checkequal(False, String("helloworld"), "startswith", String("lowo"), 3, 6)
    checkequal(True, String(""), "startswith", String(""), 0, 1)
    checkequal(True, String(""), "startswith", String(""), 0, 0)
    checkequal(False, String(""), "startswith", String(""), 1, 0)

    # test negative indices
    checkequal(True, String("hello"), "startswith", String("he"), 0, -1)
    checkequal(True, String("hello"), "startswith", String("he"), -53, -1)
    checkequal(False, String("hello"), "startswith", String("hello"), 0, -1)
    checkequal(
        False,
        String("hello"),
        "startswith",
        String("hello world"),
        -1,
        -10,
    )
    checkequal(False, String("hello"), "startswith", String("ello"), -5)
    checkequal(True, String("hello"), "startswith", String("ello"), -4)
    checkequal(False, String("hello"), "startswith", String("o"), -2)
    checkequal(True, String("hello"), "startswith", String("o"), -1)
    checkequal(True, String("hello"), "startswith", String(""), -3, -3)
    checkequal(False, String("hello"), "startswith", String("lo"), -9)

    checkraises(TypeError, String("hello"), "startswith")
    checkraises(TypeError, String("hello"), "startswith", 42)

    # test tuple arguments
    checkequal(
        True,
        String("hello"),
        "startswith",
        (String("he"), String("ha")),
    )
    checkequal(
        False,
        String("hello"),
        "startswith",
        (String("lo"), String("llo")),
    )
    checkequal(
        True,
        String("hello"),
        "startswith",
        (String("hellox"), String("hello")),
    )
    checkequal(False, String("hello"), "startswith", ())
    checkequal(
        True,
        String("helloworld"),
        "startswith",
        (String("hellowo"), String("rld"), String("lowo")),
        3,
    )
    checkequal(
        False,
        String("helloworld"),
        "startswith",
        (String("hellowo"), String("ello"), String("rld")),
        3,
    )
    checkequal(
        True,
        String("hello"),
        "startswith",
        (String("lo"), String("he")),
        0,
        -1,
    )
    checkequal(
        False,
        String("hello"),
        "startswith",
        (String("he"), String("hel")),
        0,
        1,
    )
    checkequal(
        True,
        String("hello"),
        "startswith",
        (String("he"), String("hel")),
        0,
        2,
    )

    checkraises(TypeError, String("hello"), "startswith", (42,))


def test_endswith():
    checkequal(True, String("hello"), "endswith", String("lo"))
    checkequal(False, String("hello"), "endswith", String("he"))
    checkequal(True, String("hello"), "endswith", String(""))
    checkequal(False, String("hello"), "endswith", String("hello world"))
    checkequal(False, String("helloworld"), "endswith", String("worl"))
    checkequal(True, String("helloworld"), "endswith", String("worl"), 3, 9)
    checkequal(True, String("helloworld"), "endswith", String("world"), 3, 12)
    checkequal(True, String("helloworld"), "endswith", String("lowo"), 1, 7)
    checkequal(True, String("helloworld"), "endswith", String("lowo"), 2, 7)
    checkequal(True, String("helloworld"), "endswith", String("lowo"), 3, 7)
    checkequal(False, String("helloworld"), "endswith", String("lowo"), 4, 7)
    checkequal(False, String("helloworld"), "endswith", String("lowo"), 3, 8)
    checkequal(False, String("ab"), "endswith", String("ab"), 0, 1)
    checkequal(False, String("ab"), "endswith", String("ab"), 0, 0)
    checkequal(True, String(""), "endswith", String(""), 0, 1)
    checkequal(True, String(""), "endswith", String(""), 0, 0)
    checkequal(False, String(""), "endswith", String(""), 1, 0)

    # test negative indices
    checkequal(True, String("hello"), "endswith", String("lo"), -2)
    checkequal(False, String("hello"), "endswith", String("he"), -2)
    checkequal(True, String("hello"), "endswith", String(""), -3, -3)
    checkequal(False, String("hello"), "endswith", String("hello world"), -10, -2)
    checkequal(False, String("helloworld"), "endswith", String("worl"), -6)
    checkequal(True, String("helloworld"), "endswith", String("worl"), -5, -1)
    checkequal(True, String("helloworld"), "endswith", String("worl"), -5, 9)
    checkequal(True, String("helloworld"), "endswith", String("world"), -7, 12)
    checkequal(True, String("helloworld"), "endswith", String("lowo"), -99, -3)
    checkequal(True, String("helloworld"), "endswith", String("lowo"), -8, -3)
    checkequal(True, String("helloworld"), "endswith", String("lowo"), -7, -3)
    checkequal(False, String("helloworld"), "endswith", String("lowo"), 3, -4)
    checkequal(False, String("helloworld"), "endswith", String("lowo"), -8, -2)

    checkraises(TypeError, String("hello"), "endswith")
    checkraises(TypeError, String("hello"), "endswith", 42)

    # test tuple arguments
    checkequal(
        False,
        String("hello"),
        "endswith",
        (String("he"), String("ha")),
    )
    checkequal(
        True,
        String("hello"),
        "endswith",
        (String("lo"), String("llo")),
    )
    checkequal(
        True,
        String("hello"),
        "endswith",
        (String("hellox"), String("hello")),
    )
    checkequal(False, String("hello"), "endswith", ())
    checkequal(
        True,
        String("helloworld"),
        "endswith",
        (String("hellowo"), String("rld"), String("lowo")),
        3,
    )
    checkequal(
        False,
        String("helloworld"),
        "endswith",
        (String("hellowo"), String("ello"), String("rld")),
        3,
        -1,
    )
    checkequal(
        True,
        String("hello"),
        "endswith",
        (String("hell"), String("ell")),
        0,
        -1,
    )
    checkequal(
        False,
        String("hello"),
        "endswith",
        (String("he"), String("hel")),
        0,
        1,
    )
    checkequal(
        True,
        String("hello"),
        "endswith",
        (String("he"), String("hell")),
        0,
        4,
    )

    checkraises(TypeError, String("hello"), "endswith", (42,))


def test___contains__():
    checkequal(True, String(""), "__contains__", String(""))
    checkequal(True, String("abc"), "__contains__", String(""))
    checkequal(False, String("abc"), "__contains__", String("\0"))
    checkequal(True, String("\0abc"), "__contains__", String("\0"))
    checkequal(True, String("abc\0"), "__contains__", String("\0"))
    checkequal(True, String("\0abc"), "__contains__", String("a"))
    checkequal(True, String("asdf"), "__contains__", String("asdf"))
    checkequal(False, String("asd"), "__contains__", String("asdf"))
    checkequal(False, String(""), "__contains__", String("asdf"))


def test_subscript():
    checkequal(String("a"), String("abc"), "__getitem__", 0)
    checkequal(String("c"), String("abc"), "__getitem__", -1)
    checkequal(String("a"), String("abc"), "__getitem__", 0)
    checkequal(String("abc"), String("abc"), "__getitem__", slice(0, 3))
    checkequal(String("abc"), String("abc"), "__getitem__", slice(0, 1000))
    checkequal(String("a"), String("abc"), "__getitem__", slice(0, 1))
    checkequal(String(""), String("abc"), "__getitem__", slice(0, 0))

    checkraises(TypeError, String("abc"), "__getitem__", String("def"))


def test_slice():
    checkequal(String("abc"), String("abc"), "__getitem__", slice(0, 1000))
    checkequal(String("abc"), String("abc"), "__getitem__", slice(0, 3))
    checkequal(String("ab"), String("abc"), "__getitem__", slice(0, 2))
    checkequal(String("bc"), String("abc"), "__getitem__", slice(1, 3))
    checkequal(String("b"), String("abc"), "__getitem__", slice(1, 2))
    checkequal(String(""), String("abc"), "__getitem__", slice(2, 2))
    checkequal(String(""), String("abc"), "__getitem__", slice(1000, 1000))
    checkequal(String(""), String("abc"), "__getitem__", slice(2000, 1000))
    checkequal(String(""), String("abc"), "__getitem__", slice(2, 1))

    checkraises(TypeError, String("abc"), "__getitem__", String("def"))


@pytest.mark.slow
def test_extended_getslice():
    # Test extended slicing by comparing with list slicing.
    s = string.ascii_letters + string.digits
    indices = (0, None, 1, 3, 41, sys.maxsize, -1, -2, -37)
    for start in indices:
        for stop in indices:
            # Skip step 0 (invalid)
            for step in indices[1:]:
                L = list(s)[start:stop:step]
                checkequal(
                    String("").join(L),
                    s,
                    "__getitem__",
                    slice(start, stop, step),
                )


def test_mul():
    checkequal(String(""), String("abc"), "__mul__", -1)
    checkequal(String(""), String("abc"), "__mul__", 0)
    checkequal(String("abc"), String("abc"), "__mul__", 1)
    checkequal(String("abcabcabc"), String("abc"), "__mul__", 3)
    checkraises(TypeError, String("abc"), "__mul__")
    checkraises(TypeError, String("abc"), "__mul__", String(""))
    # XXX: on a 64-bit system, this doesn't raise an overflow error,
    # but either raises a MemoryError, or succeeds (if you have 54TiB)
    # checkraises(OverflowError, 10000*'abc', '__mul__', 2000000000)


@pytest.mark.slow
def test_join():
    checkequal(
        String("a b c d"),
        String(" "),
        "join",
        [
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ],
    )
    checkequal(
        String("abcd"),
        String(""),
        "join",
        (
            String("a"),
            String("b"),
            String("c"),
            String("d"),
        ),
    )
    checkequal(
        String("bd"),
        String(""),
        "join",
        (String(""), String("b"), String(""), String("d")),
    )
    checkequal(
        String("ac"),
        String(""),
        "join",
        (String("a"), String(""), String("c"), String("")),
    )
    checkequal(String("w x y z"), String(" "), "join", Sequence())
    checkequal(String("abc"), String("a"), "join", (String("abc"),))
    checkequal(
        String("z"),
        String("a"),
        "join",
        List([String("z")]),
    )
    checkequal(
        String("a.b.c"),
        String("."),
        "join",
        [String("a"), String("b"), String("c")],
    )
    checkraises(
        TypeError,
        String("."),
        "join",
        [String("a"), String("b"), 3],
    )
    for i in [5, 25, 125]:
        checkequal(
            (((String("a") * i) + String("-")) * i)[:-1],
            String("-"),
            "join",
            [String("a") * i] * i,
        )
        checkequal(
            (((String("a") * i) + String("-")) * i)[:-1],
            String("-"),
            "join",
            (String("a") * i,) * i,
        )

    checkequal(String("a b c"), String(" "), "join", BadSeq2())

    checkraises(TypeError, String(" "), "join")
    checkraises(TypeError, String(" "), "join", None)
    checkraises(TypeError, String(" "), "join", 7)
    checkraises(TypeError, String(" "), "join", [1, 2, bytes()])


def test_formatting():
    checkequal(
        String("+hello+"),
        String("+%s+"),
        "__mod__",
        String("hello"),
    )
    checkequal(String("+10+"), String("+%d+"), "__mod__", 10)
    checkequal(String("a"), String("%c"), "__mod__", String("a"))
    checkequal(String("a"), String("%c"), "__mod__", String("a"))
    checkequal(String("$"), String("%c"), "__mod__", 36)
    checkequal(String("10"), String("%d"), "__mod__", 10)
    checkequal(String("\x7f"), String("%c"), "__mod__", 0x7F)

    for ordinal in (-100, 0x200000):
        # unicode raises ValueError, str raises OverflowError
        checkraises((ValueError, OverflowError), String("%c"), "__mod__", ordinal)

    longvalue = sys.maxsize + 10
    slongvalue = str(longvalue)
    checkequal(String(" 42"), String("%3ld"), "__mod__", 42)
    checkequal(String("42"), String("%d"), "__mod__", 42.0)
    checkequal(slongvalue, String("%d"), "__mod__", longvalue)
    checkcall(String("%d"), "__mod__", float(longvalue))
    checkequal(String("0042.00"), String("%07.2f"), "__mod__", 42)
    checkequal(String("0042.00"), String("%07.2F"), "__mod__", 42)

    checkraises(TypeError, String("abc"), "__mod__")
    checkraises(TypeError, String("%(foo)s"), "__mod__", 42)
    checkraises(TypeError, String("%s%s"), "__mod__", (42,))
    checkraises(TypeError, String("%c"), "__mod__", (None,))
    checkraises(ValueError, String("%(foo"), "__mod__", {})
    checkraises(
        TypeError,
        String("%(foo)s %(bar)s"),
        "__mod__",
        (String("foo"), 42),
    )
    checkraises(TypeError, String("%d"), "__mod__", String("42"))  # not numeric
    checkraises(
        TypeError, String("%d"), "__mod__", (42 + 0j)
    )  # no int conversion provided

    # argument names with properly nested brackets are supported
    checkequal(
        String("bar"),
        String("%((foo))s"),
        "__mod__",
        {String("(foo)"): String("bar")},
    )

    # 100 is a magic number in PyUnicode_Format, this forces a resize
    checkequal(
        103 * String("a") + String("x"),
        String("%sx"),
        "__mod__",
        103 * String("a"),
    )

    checkraises(
        TypeError,
        String("%*s"),
        "__mod__",
        (String("foo"), String("bar")),
    )
    checkraises(TypeError, String("%10.*f"), "__mod__", (String("foo"), 42.0))
    checkraises(ValueError, String("%10"), "__mod__", (42,))

    # Outrageously large width or precision should raise ValueError.
    checkraises(ValueError, String("%%%df") % (2 ** 64), "__mod__", (3.2))
    checkraises(ValueError, String("%%.%df") % (2 ** 64), "__mod__", (3.2))
    checkraises(
        OverflowError,
        String("%*s"),
        "__mod__",
        (sys.maxsize + 1, String("")),
    )
    checkraises(OverflowError, String("%.*f"), "__mod__", (sys.maxsize + 1, 1.0 / 7))

    class X(object):
        pass

    checkraises(TypeError, String("abc"), "__mod__", X())


@support.cpython_only
def test_formatting_c_limits():
    # third party
    from _testcapi import INT_MAX
    from _testcapi import PY_SSIZE_T_MAX
    from _testcapi import UINT_MAX

    SIZE_MAX = (1 << (PY_SSIZE_T_MAX.bit_length() + 1)) - 1
    checkraises(
        OverflowError,
        String("%*s"),
        "__mod__",
        (PY_SSIZE_T_MAX + 1, String("")),
    )
    checkraises(OverflowError, String("%.*f"), "__mod__", (INT_MAX + 1, 1.0 / 7))
    # Issue 15989
    checkraises(
        OverflowError,
        String("%*s"),
        "__mod__",
        (SIZE_MAX + 1, String("")),
    )
    checkraises(OverflowError, String("%.*f"), "__mod__", (UINT_MAX + 1, 1.0 / 7))


@pytest.mark.slow
def test_floatformatting():
    # float formatting
    for prec in range(100):
        format = String("%%.%if") % prec
        value = 0.01
        for x in range(60):
            value = value * 3.14159265359 / 3.0 * 10.0
            checkcall(format, "__mod__", value)


def test_inplace_rewrites():
    # Check that strings don't copy and modify cached single-character strings
    checkequal(String("a"), String("A"), "lower")
    checkequal(True, String("A"), "isupper")
    checkequal(String("A"), String("a"), "upper")
    checkequal(True, String("a"), "islower")

    checkequal(
        String("a"),
        String("A"),
        "replace",
        String("A"),
        String("a"),
    )
    checkequal(True, String("A"), "isupper")

    checkequal(String("A"), String("a"), "capitalize")
    checkequal(True, String("a"), "islower")

    checkequal(String("A"), String("a"), "swapcase")
    checkequal(True, String("a"), "islower")

    checkequal(String("A"), String("a"), "title")
    checkequal(True, String("a"), "islower")


def test_partition():
    checkequal(
        (
            String("this is the par"),
            String("ti"),
            String("tion method"),
        ),
        String("this is the partition method"),
        "partition",
        String("ti"),
    )

    # from raymond's original specification
    S = String("http://www.python.org")
    checkequal(
        (String("http"), String("://"), String("www.python.org")),
        S,
        "partition",
        String("://"),
    )
    checkequal(
        (String("http://www.python.org"), String(""), String("")),
        S,
        "partition",
        String("?"),
    )
    checkequal(
        (String(""), String("http://"), String("www.python.org")),
        S,
        "partition",
        String("http://"),
    )
    checkequal(
        (String("http://www.python."), String("org"), String("")),
        S,
        "partition",
        String("org"),
    )

    checkraises(ValueError, S, "partition", String(""))
    checkraises(TypeError, S, "partition", None)


def test_rpartition():
    checkequal(
        (
            String("this is the rparti"),
            String("ti"),
            String("on method"),
        ),
        String("this is the rpartition method"),
        "rpartition",
        String("ti"),
    )

    # from raymond's original specification
    S = String("http://www.python.org")
    checkequal(
        (String("http"), String("://"), String("www.python.org")),
        S,
        "rpartition",
        String("://"),
    )
    checkequal(
        (String(""), String(""), String("http://www.python.org")),
        S,
        "rpartition",
        String("?"),
    )
    checkequal(
        (String(""), String("http://"), String("www.python.org")),
        S,
        "rpartition",
        String("http://"),
    )
    checkequal(
        (String("http://www.python."), String("org"), String("")),
        S,
        "rpartition",
        String("org"),
    )

    checkraises(ValueError, S, "rpartition", String(""))
    checkraises(TypeError, S, "rpartition", None)


def test_none_arguments():
    # issue 11828
    s = String("hello")
    checkequal(2, s, "find", String("l"), None)
    checkequal(3, s, "find", String("l"), -2, None)
    checkequal(2, s, "find", String("l"), None, -2)
    checkequal(0, s, "find", String("h"), None, None)

    checkequal(3, s, "rfind", String("l"), None)
    checkequal(3, s, "rfind", String("l"), -2, None)
    checkequal(2, s, "rfind", String("l"), None, -2)
    checkequal(0, s, "rfind", String("h"), None, None)

    checkequal(2, s, "index", String("l"), None)
    checkequal(3, s, "index", String("l"), -2, None)
    checkequal(2, s, "index", String("l"), None, -2)
    checkequal(0, s, "index", String("h"), None, None)

    checkequal(3, s, "rindex", String("l"), None)
    checkequal(3, s, "rindex", String("l"), -2, None)
    checkequal(2, s, "rindex", String("l"), None, -2)
    checkequal(0, s, "rindex", String("h"), None, None)

    checkequal(2, s, "count", String("l"), None)
    checkequal(1, s, "count", String("l"), -2, None)
    checkequal(1, s, "count", String("l"), None, -2)
    checkequal(0, s, "count", String("x"), None, None)

    checkequal(True, s, "endswith", String("o"), None)
    checkequal(True, s, "endswith", String("lo"), -2, None)
    checkequal(True, s, "endswith", String("l"), None, -2)
    checkequal(False, s, "endswith", String("x"), None, None)

    checkequal(True, s, "startswith", String("h"), None)
    checkequal(True, s, "startswith", String("l"), -2, None)
    checkequal(True, s, "startswith", String("h"), None, -2)
    checkequal(False, s, "startswith", String("x"), None, None)



# stdlib
import json
from typing import Dict
from typing import Optional
from typing import Tuple

# third party
import requests

# syft relative
from ...core.common.message import SyftMessage
from ...core.common.serde.serialize import _serialize
from ...proto.core.node.common.metadata_pb2 import Metadata as Metadata_PB
from ..connections.http_connection import HTTPConnection


class GridHTTPConnection(HTTPConnection):
    LOGIN_ROUTE = "/users/login"
    SYFT_ROUTE = "/pysyft"

    def __init__(self, url: str) -> None:
        self.base_url = url
        self.session_token: Optional[Dict[str, str]] = None

    def _send_msg(self, msg: SyftMessage) -> requests.Response:
        """
        Serializes Syft messages in json format and send it using HTTP protocol.
        NOTE: Auxiliary method to avoid code duplication and modularity.
        :return: returns requests.Response object containing a JSON serialized
        SyftMessage
        :rtype: requests.Response
        """

        header = {}

        if self.session_token:
            header["token"] = self.session_token

        header["Content-Type"] = "application/octet-stream"  # type: ignore

        # Perform HTTP request using base_url as a root address
        msg_bytes: bytes = _serialize(obj=msg, to_bytes=True)  # type: ignore
        r = requests.post(
            url=self.base_url + GridHTTPConnection.SYFT_ROUTE,
            data=msg_bytes,
            headers=header,
        )

        # Return request's response object
        # r.text provides the response body as a str
        return r

    def login(self, credentials: Dict) -> Tuple:
        # Login request
        response = requests.post(
            url=self.base_url + GridHTTPConnection.LOGIN_ROUTE, json=credentials
        )

        # Response
        content = json.loads(response.text)

        # If fail
        if response.status_code != requests.codes.ok:
            raise Exception(content["error"])

        metadata = content["metadata"].encode("ISO-8859-1")
        metadata_pb = Metadata_PB()
        metadata_pb.ParseFromString(metadata)

        # If success
        # Save session token
        self.session_token = content["token"]

        # Return node metadata / user private key
        return (metadata_pb, content["key"])

    def _get_metadata(self) -> Tuple:
        """Request Node's metadata
        :return: returns node metadata
        :rtype: str of bytes
        """
        response = requests.get(self.base_url + "/metadata")
        content = json.loads(response.text)

        metadata = content["metadata"].encode("ISO-8859-1")
        metadata_pb = Metadata_PB()
        metadata_pb.ParseFromString(metadata)

        return metadata_pb



# stdlib
from typing import Union

# third party
from nacl.signing import VerifyKey

# syft relative
from ....common.message import EventualSyftMessageWithoutReply
from ....common.message import ImmediateSyftMessageWithReply
from ....common.message import ImmediateSyftMessageWithoutReply
from ....common.message import SyftMessage
from ...abstract.node import AbstractNode


class Action(SyftMessage):
    ""

    def execute_action(
        self, node: AbstractNode, verify_key: VerifyKey
    ) -> Union[SyftMessage, None]:
        raise NotImplementedError


class ImmediateActionWithoutReply(Action, ImmediateSyftMessageWithoutReply):
    ""

    def execute_action(self, node: AbstractNode, verify_key: VerifyKey) -> None:
        raise NotImplementedError


class EventualActionWithoutReply(Action, EventualSyftMessageWithoutReply):
    ""

    def execute_action(self, node: AbstractNode, verify_key: VerifyKey) -> None:
        raise NotImplementedError


class ImmediateActionWithReply(Action, ImmediateSyftMessageWithReply):
    ""

    def execute_action(
        self, node: AbstractNode, verify_key: VerifyKey
    ) -> ImmediateSyftMessageWithoutReply:
        raise NotImplementedError



# stdlib
from typing import Any
from typing import Type
from typing import Union

# third party
from nacl.signing import SigningKey

# syft relative
from ...core.common.message import ImmediateSyftMessageWithReply
from ...core.common.message import ImmediateSyftMessageWithoutReply
from ...core.common.message import SignedImmediateSyftMessageWithReply
from ...core.common.message import SignedImmediateSyftMessageWithoutReply
from ...core.common.message import SyftMessage
from ...core.io.address import Address
from ...core.io.route import SoloRoute
from ...core.node.network.client import NetworkClient
from ..services.signaling_service import RegisterNewPeerMessage


class SignalingClient(object):
    def __init__(
        self, url: str, conn_type: type, client_type: Type[NetworkClient]
    ) -> None:
        # Load an Signing Key instance
        signing_key = SigningKey.generate()
        verify_key = signing_key.verify_key

        # Use Signaling Server metadata
        # to build client route
        conn = conn_type(url=url)
        (
            spec_location,
            name,
            client_id,
        ) = client_type.deserialize_client_metadata_from_node(
            metadata=conn._get_metadata()
        )

        # Create a new Solo Route using the selected connection type
        route = SoloRoute(destination=spec_location, connection=conn)

        # Create a new signaling client using the selected client type
        signaling_client = client_type(
            network=spec_location,
            name=name,
            routes=[route],
            signing_key=signing_key,
            verify_key=verify_key,
        )

        self.__client = signaling_client
        self.__register()

    @property
    def address(self) -> Address:
        return self.__client.address

    def __register(self) -> None:
        _response = self.__client.send_immediate_msg_with_reply(
            msg=RegisterNewPeerMessage(
                address=self.__client.address, reply_to=self.__client.address
            )
        )
        self.duet_id = _response.peer_id  # type: ignore

    def send_immediate_msg_with_reply(
        self,
        msg: Union[SignedImmediateSyftMessageWithReply, ImmediateSyftMessageWithReply],
    ) -> SyftMessage:
        return self.__client.send_immediate_msg_with_reply(msg=msg)

    def send_immediate_msg_without_reply(
        self,
        msg: Union[
            SignedImmediateSyftMessageWithoutReply, ImmediateSyftMessageWithoutReply
        ],
    ) -> None:
        self.__client.send_immediate_msg_without_reply(msg=msg)

    def send_eventual_msg_without_reply(self, msg: Any) -> None:
        self.__client.send_eventual_msg_without_reply(msg=msg)



# stdlib
from typing import Any
from typing import Callable
from typing import Type

# third party
from pandas import DataFrame

# syft relative
from ....core.common.serde.deserialize import _deserialize
from ....core.pointer.pointer import Pointer
from ....proto.core.io.address_pb2 import Address as Address_PB
from ...messages.infra_messages import CreateWorkerMessage
from ...messages.infra_messages import DeleteWorkerMessage
from ...messages.infra_messages import GetWorkerInstanceTypesMessage
from ...messages.infra_messages import GetWorkerMessage
from ...messages.infra_messages import GetWorkersMessage
from ...messages.infra_messages import UpdateWorkerMessage
from ...messages.transfer_messages import SaveObjectMessage
from .request_api import GridRequestAPI


class WorkerRequestAPI(GridRequestAPI):
    response_key = "worker"

    def __init__(self, send: Callable, domain_client: Any):
        super().__init__(
            create_msg=CreateWorkerMessage,
            get_msg=GetWorkerMessage,
            get_all_msg=GetWorkersMessage,
            update_msg=UpdateWorkerMessage,
            delete_msg=DeleteWorkerMessage,
            send=send,
            response_key=WorkerRequestAPI.response_key,
        )

        self.domain_client = domain_client

    def instance_type(self, pandas: bool = False) -> Any:
        result = self.send(grid_msg=GetWorkerInstanceTypesMessage)
        if pandas:
            max_size = len(min(result.values()))  # Why min/max functions were switched?
            for key in result.keys():
                empty_cells = max_size - len(result[key])
                while empty_cells:
                    result[key].append("")
                    empty_cells -= 1
            result = DataFrame(result)
        return result

    def __getitem__(self, key: int) -> object:
        return self.get(worker_id=key)

    def __delitem__(self, key: int) -> None:
        self.delete(worker_id=key)

    def to_obj(self, result: Any) -> Any:
        _raw_worker = super().to_obj(result)
        _raw_addr = _raw_worker.syft_address.encode("ISO-8859-1")

        addr_pb = Address_PB()
        addr_pb.ParseFromString(_raw_addr)

        _worker_obj = self.domain_client.__class__(
            credentials={},
            url=self.domain_client.conn.base_url,
            conn_type=self.domain_client.conn.__class__,
            client_type=self.domain_client.client_type,
        )
        _worker_obj.proxy_address = _deserialize(blob=addr_pb)
        _worker_obj.domain = _worker_obj.proxy_address.domain

        for key, value in result.items():
            try:
                setattr(_worker_obj, key, value)
            except AttributeError:
                continue

        def _save(obj_ptr: Type[Pointer]) -> None:
            _content = {
                "address": _worker_obj.address,
                "content": {
                    "uid": str(obj_ptr.id_at_location.value),
                    "domain_address": self.domain_client.conn.base_url,
                },
            }
            signed_msg = SaveObjectMessage(**_content).sign(
                signing_key=_worker_obj.signing_key
            )
            _worker_obj.send_immediate_msg_without_reply(msg=signed_msg)

        _worker_obj.save = _save
        return _worker_obj



# third party
import torch
import torch.nn as nn


class Bottle(nn.Module):
    def forward(self, input):
        if len(input.size()) <= 2:
            return super(Bottle, self).forward(input)
        size = input.size()[:2]
        out = super(Bottle, self).forward(input.view(size[0] * size[1], -1))
        return out.view(size[0], size[1], -1)


class Linear(Bottle, nn.Linear):
    pass


class Encoder(nn.Module):
    def __init__(self, config):
        super(Encoder, self).__init__()
        self.config = config
        input_size = config.d_proj if config.projection else config.d_embed
        dropout = 0 if config.n_layers == 1 else config.dp_ratio
        self.rnn = nn.LSTM(
            input_size=input_size,
            hidden_size=config.d_hidden,
            num_layers=config.n_layers,
            dropout=dropout,
            bidirectional=config.birnn,
        )

    def forward(self, inputs):
        batch_size = inputs.size()[1]
        state_shape = self.config.n_cells, batch_size, self.config.d_hidden
        h0 = c0 = inputs.new_zeros(state_shape)
        outputs, (ht, ct) = self.rnn(inputs, (h0, c0))
        return (
            ht[-1]
            if not self.config.birnn
            else ht[-2:].transpose(0, 1).contiguous().view(batch_size, -1)
        )


class SNLIClassifier(nn.Module):
    def __init__(self, config):
        super(SNLIClassifier, self).__init__()
        self.config = config
        self.embed = nn.Embedding(config.n_embed, config.d_embed)
        self.projection = Linear(config.d_embed, config.d_proj)
        self.encoder = Encoder(config)
        self.dropout = nn.Dropout(p=config.dp_ratio)
        self.relu = nn.ReLU()
        seq_in_size = 2 * config.d_hidden
        if self.config.birnn:
            seq_in_size *= 2
        lin_config = [seq_in_size] * 2
        self.out = nn.Sequential(
            Linear(*lin_config),
            self.relu,
            self.dropout,
            Linear(*lin_config),
            self.relu,
            self.dropout,
            Linear(*lin_config),
            self.relu,
            self.dropout,
            Linear(seq_in_size, config.d_out),
        )

    def forward(self, batch):
        prem_embed = self.embed(batch.premise)
        hypo_embed = self.embed(batch.hypothesis)
        if self.config.fix_emb:
            prem_embed = prem_embed.detach()
            hypo_embed = hypo_embed.detach()
        if self.config.projection:
            prem_embed = self.relu(self.projection(prem_embed))
            hypo_embed = self.relu(self.projection(hypo_embed))
        premise = self.encoder(prem_embed)
        hypothesis = self.encoder(hypo_embed)
        scores = self.out(torch.cat([premise, hypothesis], 1))
        return scores



# syft absolute
import syft as sy


def test_ordereddict_reversed() -> None:
    s = sy.lib.python.collections.OrderedDict()
    s["1"] = 1
    s["2"] = 2
    s["3"] = 3

    assert s == {"1": 1, "2": 2, "3": 3}

    reverse_keys = ["3", "2", "1"]
    for i in reversed(s):
        print(i, reverse_keys)
        assert i == reverse_keys.pop(0)



"""In this test suite, we evaluate the SpecificLocation class. For more info
on the SpecificLocation class and its purpose, please see the documentation
in the class itself.

Table of Contents:
    - INITIALIZATION: tests for various ways SpecificLocation can/can't be initialized
    - CLASS METHODS: tests for the use of SpecificLocation's class methods
    - SERDE: test for serialization and deserialization of SpecificLocation.

"""


# stdlib
import uuid

# syft absolute
import syft as sy
from syft.core.common.uid import UID
from syft.core.io.location.specific import SpecificLocation

# --------------------- INITIALIZATION ---------------------


def test_specific_location_init_without_arguments() -> None:
    """Test that SpecificLocation will self-create an ID object if none is given"""

    # init works without arguments
    loc = SpecificLocation()

    assert isinstance(loc.id, UID)


def test_specific_location_init_with_specific_id() -> None:
    """Test that SpecificLocation will use the ID you pass into the constructor"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))

    loc = SpecificLocation(id=uid)

    assert loc.id == uid


# --------------------- CLASS METHODS ---------------------


def test_compare() -> None:
    """While uses of this feature should be quite rare, we
    should be able to check whether two objects are the same
    based on their IDs being the same by default. Note that
    subclasses will undoubtedly modify this behavior with other
    __eq__ methods."""

    obj = SpecificLocation()
    obj2 = SpecificLocation()

    assert obj != obj2

    obj._id = obj2.id

    assert obj == obj2


def test_to_string() -> None:
    """Tests that SpecificLocation generates an intuitive string."""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = SpecificLocation(id=uid)
    assert str(obj) == "<SpecificLocation: fb1bb0675bb74c49becee700ab0a1514>"
    assert obj.__repr__() == "<SpecificLocation: fb1bb0675bb74c49becee700ab0a1514>"


def test_pprint() -> None:
    """Tests that SpecificLocation generates a pretty representation."""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = SpecificLocation(id=uid, name="location")
    assert obj.pprint == "📌 location (SpecificLocation)@<UID:🙍🛖>"


# --------------------- SERDE ---------------------


def test_default_serialization() -> None:
    """Tests that default SpecificLocation serialization works as expected - to Protobuf"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = SpecificLocation(id=uid, name="Test")

    blob = sy.serialize(obj, to_proto=True)

    assert sy.serialize(obj) == blob


def test_default_deserialization() -> None:
    """Tests that default SpecificLocation deserialization works as expected - from Protobuf"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = SpecificLocation(id=uid, name="Test")

    blob = SpecificLocation.get_protobuf_schema()(id=sy.serialize(uid))

    obj2 = sy.deserialize(blob=blob)
    assert obj == obj2


def test_proto_serialization() -> None:
    """Tests that default SpecificLocation serialization works as expected - to Protobuf"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = SpecificLocation(id=uid, name="Test")

    blob = SpecificLocation.get_protobuf_schema()(id=sy.serialize(uid), name="Test")

    assert sy.serialize(obj, to_proto=True) == blob
    assert sy.serialize(obj, to_proto=True) == blob
    assert sy.serialize(obj, to_proto=True) == blob


def test_proto_deserialization() -> None:
    """Tests that default SpecificLocation deserialization works as expected - from Protobuf"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = SpecificLocation(id=uid)

    blob = SpecificLocation.get_protobuf_schema()(id=sy.serialize(uid))

    obj2 = sy.deserialize(blob=blob, from_proto=True)
    assert obj == obj2


def test_binary_serialization() -> None:
    """Tests that binary SpecificLocation serializes as expected"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = SpecificLocation(id=uid, name="Test")

    blob = (
        b"\n/syft.core.io.location.specific.SpecificLocation\x12\x1a\n\x12\n\x10"
        + b"\xfb\x1b\xb0g[\xb7LI\xbe\xce\xe7\x00\xab\n\x15\x14\x12\x04Test"
    )

    assert sy.serialize(obj, to_bytes=True) == blob
    assert sy.serialize(obj, to_bytes=True) == blob
    assert sy.serialize(obj, to_bytes=True) == blob


def test_binary_deserialization() -> None:
    """Test that binary SpecificLocation deserialization works as expected"""

    blob = (
        b"\n/syft.core.io.location.specific.SpecificLocation\x12\x1a\n\x12\n\x10"
        + b"\xfb\x1b\xb0g[\xb7LI\xbe\xce\xe7\x00\xab\n\x15\x14\x12\x04Test"
    )

    obj = sy.deserialize(blob=blob, from_bytes=True)
    assert obj == SpecificLocation(
        id=UID(value=uuid.UUID(int=333779996850170035686993356951732753684)),
        name="Test",
    )



# syft relative
from .request_answer_message import RequestAnswerMessage  # noqa: F401
from .request_answer_message import RequestAnswerMessageService  # noqa: F401
from .request_answer_message import RequestAnswerResponse  # noqa: F401
from .request_message import RequestMessage  # noqa: F401
from .request_message import RequestService  # noqa: F401
from .request_message import RequestStatus  # noqa: F401



# stdlib
import math

# third party
import torch
import torch.nn as nn
import torch.nn.functional as F


class RNNModel(nn.Module):
    """Container module with an encoder, a recurrent module, and a decoder."""

    def __init__(
        self, rnn_type, ntoken, ninp, nhid, nlayers, dropout=0.5, tie_weights=False
    ):
        super(RNNModel, self).__init__()
        self.ntoken = ntoken
        self.drop = nn.Dropout(dropout)
        self.encoder = nn.Embedding(ntoken, ninp)
        if rnn_type in ["LSTM", "GRU"]:
            self.rnn = getattr(nn, rnn_type)(ninp, nhid, nlayers, dropout=dropout)
        else:
            try:
                nonlinearity = {"RNN_TANH": "tanh", "RNN_RELU": "relu"}[rnn_type]
            except KeyError:
                raise ValueError(
                    """An invalid option for `--model` was supplied,
                                 options are ['LSTM', 'GRU', 'RNN_TANH' or 'RNN_RELU']"""
                )
            self.rnn = nn.RNN(
                ninp, nhid, nlayers, nonlinearity=nonlinearity, dropout=dropout
            )
        self.decoder = nn.Linear(nhid, ntoken)

        # Optionally tie weights as in:
        # "Using the Output Embedding to Improve Language Models" (Press & Wolf 2016)
        # https://arxiv.org/abs/1608.05859
        # and
        # "Tying Word Vectors and Word Classifiers: A Loss Framework for Language Modeling" (Inan et al. 2016)
        # https://arxiv.org/abs/1611.01462
        if tie_weights:
            if nhid != ninp:
                raise ValueError(
                    "When using the tied flag, nhid must be equal to emsize"
                )
            self.decoder.weight = self.encoder.weight

        self.init_weights()

        self.rnn_type = rnn_type
        self.nhid = nhid
        self.nlayers = nlayers

    def init_weights(self):
        initrange = 0.1
        nn.init.uniform_(self.encoder.weight, -initrange, initrange)
        nn.init.zeros_(self.decoder.weight)
        nn.init.uniform_(self.decoder.weight, -initrange, initrange)

    def forward(self, input, hidden):
        emb = self.drop(self.encoder(input))
        output, hidden = self.rnn(emb, hidden)
        output = self.drop(output)
        decoded = self.decoder(output)
        decoded = decoded.view(-1, self.ntoken)
        return F.log_softmax(decoded, dim=1), hidden

    def init_hidden(self, bsz):
        weight = next(self.parameters())
        if self.rnn_type == "LSTM":
            return (
                weight.new_zeros(self.nlayers, bsz, self.nhid),
                weight.new_zeros(self.nlayers, bsz, self.nhid),
            )
        else:
            return weight.new_zeros(self.nlayers, bsz, self.nhid)


# Temporarily leave PositionalEncoding module here. Will be moved somewhere else.
class PositionalEncoding(nn.Module):
    r"""Inject some information about the relative or absolute position of the tokens
        in the sequence. The positional encodings have the same dimension as
        the embeddings, so that the two can be summed. Here, we use sine and cosine
        functions of different frequencies.
    .. math::
        \text{PosEncoder}(pos, 2i) = sin(pos/10000^(2i/d_model))
        \text{PosEncoder}(pos, 2i+1) = cos(pos/10000^(2i/d_model))
        \text{where pos is the word position and i is the embed idx)
    Args:
        d_model: the embed dim (required).
        dropout: the dropout value (default=0.1).
        max_len: the max. length of the incoming sequence (default=5000).
    Examples:
        >>> pos_encoder = PositionalEncoding(d_model)
    """

    def __init__(self, d_model, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer("pe", pe)

    def forward(self, x):
        r"""Inputs of forward function
        Args:
            x: the sequence fed to the positional encoder model (required).
        Shape:
            x: [sequence length, batch size, embed dim]
            output: [sequence length, batch size, embed dim]
        Examples:
            >>> output = pos_encoder(x)
        """

        x = x + self.pe[: x.size(0), :]
        return self.dropout(x)


class TransformerModel(nn.Module):
    """Container module with an encoder, a recurrent or transformer module, and a decoder."""

    def __init__(self, ntoken, ninp, nhead, nhid, nlayers, dropout=0.5):
        super(TransformerModel, self).__init__()
        try:
            # third party
            from torch.nn import TransformerEncoder
            from torch.nn import TransformerEncoderLayer
        except:
            raise ImportError(
                "TransformerEncoder module does not exist in PyTorch 1.1 or lower."
            )
        self.model_type = "Transformer"
        self.src_mask = None
        self.pos_encoder = PositionalEncoding(ninp, dropout)
        encoder_layers = TransformerEncoderLayer(ninp, nhead, nhid, dropout)
        self.transformer_encoder = TransformerEncoder(encoder_layers, nlayers)
        self.encoder = nn.Embedding(ntoken, ninp)
        self.ninp = ninp
        self.decoder = nn.Linear(ninp, ntoken)

        self.init_weights()

    def _generate_square_subsequent_mask(self, sz):
        mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
        mask = (
            mask.float()
            .masked_fill(mask == 0, float("-inf"))
            .masked_fill(mask == 1, float(0.0))
        )
        return mask

    def init_weights(self):
        initrange = 0.1
        nn.init.uniform_(self.encoder.weight, -initrange, initrange)
        nn.init.zeros_(self.decoder.weight)
        nn.init.uniform_(self.decoder.weight, -initrange, initrange)

    def forward(self, src, has_mask=True):
        if has_mask:
            device = src.device
            if self.src_mask is None or self.src_mask.size(0) != len(src):
                mask = self._generate_square_subsequent_mask(len(src)).to(device)
                self.src_mask = mask
        else:
            self.src_mask = None

        src = self.encoder(src) * math.sqrt(self.ninp)
        src = self.pos_encoder(src)
        output = self.transformer_encoder(src, self.src_mask)
        output = self.decoder(output)
        return F.log_softmax(output, dim=-1)



# stdlib
from typing import Any
from typing import List
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# syft relative
from .....logger import traceback_and_raise
from ...abstract.node import AbstractNode


class NodeService:
    @staticmethod
    def message_handler_types() -> List[Type[Any]]:
        traceback_and_raise(NotImplementedError)

    @classmethod
    def class_name(cls) -> str:
        return str(cls.__name__)

    @classmethod
    def icon(cls) -> str:
        return "⚙️ "

    @classmethod
    def pprint(cls) -> str:
        return f"{cls.icon()} ({cls.class_name()})"

    @staticmethod
    def process(
        node: AbstractNode, msg: Any, verify_key: Optional[VerifyKey] = None
    ) -> Any:
        traceback_and_raise(NotImplementedError)


class ImmediateNodeService(NodeService):
    """A service for messages which should be immediately executed"""

    pass


class EventualNodeService(NodeService):
    """A service for messages which need not be immediately executed
    but which can be executed at the worker's convenience"""

    pass


class ImmediateNodeServiceWithReply(ImmediateNodeService):
    pass


class EventualNodeServiceWithoutReply(NodeService):
    pass


class ImmediateNodeServiceWithoutReply(NodeService):
    pass


class SignedNodeServiceWithReply(ImmediateNodeService):
    pass


class SignedNodeServiceWithoutReply(ImmediateNodeService):
    pass



# future
from __future__ import print_function

# stdlib
import argparse

# third party
from PIL import Image
import numpy as np
import torch
from torchvision.transforms import ToTensor

# Training settings
parser = argparse.ArgumentParser(description="PyTorch Super Res Example")
parser.add_argument("--input_image", type=str, required=True, help="input image to use")
parser.add_argument("--model", type=str, required=True, help="model file to use")
parser.add_argument(
    "--output_filename", type=str, help="where to save the output image"
)
parser.add_argument("--cuda", action="store_true", help="use cuda")
opt = parser.parse_args()

print(opt)
img = Image.open(opt.input_image).convert("YCbCr")
y, cb, cr = img.split()

model = torch.load(opt.model)
img_to_tensor = ToTensor()
input = img_to_tensor(y).view(1, -1, y.size[1], y.size[0])

if opt.cuda:
    model = model.cuda()
    input = input.cuda()

out = model(input)
out = out.cpu()
out_img_y = out[0].detach().numpy()
out_img_y *= 255.0
out_img_y = out_img_y.clip(0, 255)
out_img_y = Image.fromarray(np.uint8(out_img_y[0]), mode="L")

out_img_cb = cb.resize(out_img_y.size, Image.BICUBIC)
out_img_cr = cr.resize(out_img_y.size, Image.BICUBIC)
out_img = Image.merge("YCbCr", [out_img_y, out_img_cb, out_img_cr]).convert("RGB")

out_img.save(opt.output_filename)
print("output image saved to ", opt.output_filename)



# stdlib
from io import open
import os

# third party
import torch


class Dictionary(object):
    def __init__(self):
        self.word2idx = {}
        self.idx2word = []

    def add_word(self, word):
        if word not in self.word2idx:
            self.idx2word.append(word)
            self.word2idx[word] = len(self.idx2word) - 1
        return self.word2idx[word]

    def __len__(self):
        return len(self.idx2word)


class Corpus(object):
    def __init__(self, path):
        self.dictionary = Dictionary()
        self.train = self.tokenize(os.path.join(path, "train.txt"))
        self.valid = self.tokenize(os.path.join(path, "valid.txt"))
        self.test = self.tokenize(os.path.join(path, "test.txt"))

    def tokenize(self, path):
        """Tokenizes a text file."""
        assert os.path.exists(path)
        # Add words to the dictionary
        with open(path, "r", encoding="utf8") as f:
            for line in f:
                words = line.split() + ["<eos>"]
                for word in words:
                    self.dictionary.add_word(word)

        # Tokenize file content
        with open(path, "r", encoding="utf8") as f:
            idss = []
            for line in f:
                words = line.split() + ["<eos>"]
                ids = []
                for word in words:
                    ids.append(self.dictionary.word2idx[word])
                idss.append(torch.tensor(ids).type(torch.int64))
            ids = torch.cat(idss)

        return ids



# stdlib
from typing import Any
from typing import Callable

# syft relative
from ...messages.group_messages import CreateGroupMessage
from ...messages.group_messages import DeleteGroupMessage
from ...messages.group_messages import GetGroupMessage
from ...messages.group_messages import GetGroupsMessage
from ...messages.group_messages import UpdateGroupMessage
from .request_api import GridRequestAPI


class GroupRequestAPI(GridRequestAPI):
    response_key = "group"

    def __init__(self, send: Callable):
        super().__init__(
            create_msg=CreateGroupMessage,
            get_msg=GetGroupMessage,
            get_all_msg=GetGroupsMessage,
            update_msg=UpdateGroupMessage,
            delete_msg=DeleteGroupMessage,
            send=send,
            response_key=GroupRequestAPI.response_key,
        )

    def __getitem__(self, key: int) -> Any:
        return self.get(group_id=key)

    def __delitem__(self, key: int) -> Any:
        self.delete(group_id=key)



# flake8: noqa
"""
File copied from cpython test suite:
https://github.com/python/cpython/blob/3.9/Lib/test/test_ordered_dict.py

Sanity tests for OrderedDict
"""

# stdlib
from collections.abc import MutableMapping
import copy
import gc
import pickle
from random import randrange
from random import shuffle
import sys
import weakref

# third party
import pytest

# syft absolute
from syft.core.common.uid import UID
from syft.lib.python import SyNone
from syft.lib.python.collections import OrderedDict as SyOrderedDict


def assertEqual(left, right):
    assert left == right


def assertNotEqual(left, right):
    assert left != right


def assertNotIn(key, container):
    assert key not in container


def assertRaises(exc, obj, methodname, *args):
    with pytest.raises(exc) as e_info:
        getattr(obj, methodname)(*args)
    assert str(e_info) != ""


def test_init():
    OrderedDict = SyOrderedDict
    with pytest.raises(TypeError):
        OrderedDict([("a", 1), ("b", 2)], None)  # too many args
    pairs = [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5)]
    assertEqual(sorted(OrderedDict(dict(pairs)).items()), pairs)  # dict input
    assertEqual(sorted(OrderedDict(**dict(pairs)).items()), pairs)  # kwds input
    assertEqual(list(OrderedDict(pairs).items()), pairs)  # pairs input
    assertEqual(
        list(OrderedDict([("a", 1), ("b", 2), ("c", 9), ("d", 4)], c=3, e=5).items()),
        pairs,
    )  # mixed input

    # make sure no positional args conflict with possible kwdargs
    assertEqual(list(OrderedDict(other=42).items()), [("other", 42)])
    assertRaises(TypeError, OrderedDict, 42)
    assertRaises(TypeError, OrderedDict, (), ())

    # Make sure that direct calls to __init__ do not clear previous contents
    d = OrderedDict([("a", 1), ("b", 2), ("c", 3), ("d", 44), ("e", 55)])
    d.__init__([("e", 5), ("f", 6)], g=7, d=4)
    assertEqual(
        list(d.items()),
        [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5), ("f", 6), ("g", 7)],
    )


def test_468():
    OrderedDict = SyOrderedDict
    items = [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5), ("f", 6), ("g", 7)]
    shuffle(items)
    argdict = OrderedDict(items)
    unpacked = {}
    for item in argdict:
        unpacked[str(item)] = argdict[item]
    d = OrderedDict(unpacked)
    assertEqual(list(d.items()), items)


def test_update():
    OrderedDict = SyOrderedDict
    with pytest.raises(TypeError):
        OrderedDict().update([("a", 1), ("b", 2)], None)  # too many args
    pairs = [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5)]
    od = OrderedDict()
    od.update(dict(pairs))
    assertEqual(sorted(od.items()), pairs)  # dict input
    od = OrderedDict()
    od.update(**dict(pairs))
    assertEqual(sorted(od.items()), pairs)  # kwds input
    od = OrderedDict()
    od.update(pairs)
    assertEqual(list(od.items()), pairs)  # pairs input
    od = OrderedDict()
    od.update([("a", 1), ("b", 2), ("c", 9), ("d", 4)], c=3, e=5)
    assertEqual(list(od.items()), pairs)  # mixed input

    # Issue 9137: Named argument called 'other' or ''
    # shouldn't be treated specially.
    od = OrderedDict()

    od = OrderedDict()
    od.update(other={})
    assertEqual(list(od.items()), [("other", {})])

    # Make sure that direct calls to update do not clear previous contents
    # add that updates items are not moved to the end
    d = OrderedDict([("a", 1), ("b", 2), ("c", 3), ("d", 44), ("e", 55)])
    d.update([("e", 5), ("f", 6)], g=7, d=4)
    assertEqual(
        list(d.items()),
        [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5), ("f", 6), ("g", 7)],
    )

    pytest.raises(TypeError, OrderedDict().update, 42)
    pytest.raises(TypeError, OrderedDict().update, (), ())

    d = OrderedDict(
        [("a", 1), ("b", 2), ("c", 3), ("d", 44), ("e", 55)],
        _id=UID.from_string(value="{12345678-1234-5678-1234-567812345678}"),
    )
    assert d.id.__eq__(UID.from_string(value="{12345678-1234-5678-1234-567812345678}"))


def test_init_calls():
    calls = []

    class Spam:
        def keys(self):
            calls.append("keys")
            return ()

        def items(self):
            calls.append("items")
            return ()

    SyOrderedDict(Spam())
    assertEqual(calls, ["keys"])


def test_FromKeys():
    OrderedDict = SyOrderedDict
    od = OrderedDict.FromKeys("abc")
    assertEqual(list(od.items()), [(c, None) for c in "abc"])
    od = OrderedDict.FromKeys("abc", value=None)
    assertEqual(list(od.items()), [(c, None) for c in "abc"])
    od = OrderedDict.FromKeys("abc", value=0)
    assertEqual(list(od.items()), [(c, 0) for c in "abc"])


def test_abc():
    OrderedDict = SyOrderedDict
    assert isinstance(OrderedDict(), MutableMapping)
    assert issubclass(OrderedDict, MutableMapping)


def test_clear():
    OrderedDict = SyOrderedDict
    pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
    shuffle(pairs)
    od = OrderedDict(pairs)
    assertEqual(len(od), len(pairs))
    od.clear()
    assertEqual(len(od), 0)


def test_delitem():
    OrderedDict = SyOrderedDict
    pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
    od = OrderedDict(pairs)
    del od["a"]
    assertNotIn("a", od)
    with pytest.raises(KeyError):
        del od["a"]
    assertEqual(list(od.items()), pairs[:2] + pairs[3:])


def test_setitem():
    OrderedDict = SyOrderedDict
    od = OrderedDict([("d", 1), ("b", 2), ("c", 3), ("a", 4), ("e", 5)])
    od["c"] = 10  # existing element
    od["f"] = 20  # new element
    assertEqual(
        list(od.items()), [("d", 1), ("b", 2), ("c", 10), ("a", 4), ("e", 5), ("f", 20)]
    )


def test_iterators():
    OrderedDict = SyOrderedDict
    pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
    shuffle(pairs)
    od = OrderedDict(pairs)
    assertEqual(list(od), [t[0] for t in pairs])
    assertEqual(list(od.keys()), [t[0] for t in pairs])
    assertEqual(list(od.values()), [t[1] for t in pairs])
    assertEqual(list(od.items()), pairs)
    assertEqual(list(reversed(od)), [t[0] for t in reversed(pairs)])
    assertEqual(list(reversed(list(od.keys()))), [t[0] for t in reversed(pairs)])
    assertEqual(list(reversed(list(od.values()))), [t[1] for t in reversed(pairs)])
    assertEqual(list(reversed(list(od.items()))), list(reversed(pairs)))


def test_detect_deletion_during_iteration():
    OrderedDict = SyOrderedDict
    od = OrderedDict.FromKeys("abc")
    it = iter(od)
    key = next(it)
    del od[key]
    with pytest.raises(Exception):
        # Note, the exact exception raised is not guaranteed
        # The only guarantee that the next() will not succeed
        next(it)


def test_sorted_iterators():
    OrderedDict = SyOrderedDict
    with pytest.raises(TypeError):
        OrderedDict([("a", 1), ("b", 2)], None)
    pairs = [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5)]
    od = OrderedDict(pairs)
    assertEqual(sorted(od), [t[0] for t in pairs])
    assertEqual(sorted(od.keys()), [t[0] for t in pairs])
    assertEqual(sorted(od.values()), [t[1] for t in pairs])
    assertEqual(sorted(od.items()), pairs)
    assertEqual(sorted(reversed(od)), sorted([t[0] for t in reversed(pairs)]))


def test_iterators_empty():
    OrderedDict = SyOrderedDict
    od = OrderedDict()
    empty = []
    assertEqual(list(od), empty)
    assertEqual(list(od.keys()), empty)
    assertEqual(list(od.values()), empty)
    assertEqual(list(od.items()), empty)
    assertEqual(list(reversed(od)), empty)
    assertEqual(list(reversed(list(od.keys()))), empty)
    assertEqual(list(reversed(list(od.values()))), empty)
    assertEqual(list(reversed(list(od.items()))), empty)


def test_popitem():
    OrderedDict = SyOrderedDict
    pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
    shuffle(pairs)
    od = OrderedDict(pairs)
    while pairs:
        assertEqual(od.popitem(), pairs.pop())
    with pytest.raises(KeyError):
        od.popitem()
    assertEqual(len(od), 0)


def test_popitem_last():
    OrderedDict = SyOrderedDict
    pairs = [(i, i) for i in range(30)]

    obj = OrderedDict(pairs)
    for i in range(8):
        obj.popitem(True)
    obj.popitem(True)
    obj.popitem(last=True)
    assertEqual(len(obj), 20)


def test_pop():
    OrderedDict = SyOrderedDict
    pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
    shuffle(pairs)
    od = OrderedDict(pairs)
    shuffle(pairs)
    while pairs:
        k, v = pairs.pop()
        assertEqual(od.pop(k), v)
    with pytest.raises(KeyError):
        od.pop("xyz")
    assertEqual(len(od), 0)
    assertEqual(od.pop(k, 12345), 12345)

    # make sure pop still works when __missing__ is defined
    class Missing(OrderedDict):
        def __missing__(self, key):
            return 0

    m = Missing(a=1)
    assertEqual(m.pop("b", 5), 5)
    assertEqual(m.pop("a", 6), 1)
    assertEqual(m.pop("a", 6), 6)
    assertEqual(m.pop("a", default=6), 6)
    with pytest.raises(KeyError):
        m.pop("a")


def test_equality():
    OrderedDict = SyOrderedDict
    pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
    shuffle(pairs)
    od1 = OrderedDict(pairs)
    od2 = OrderedDict(pairs)
    assertEqual(od1, od2)  # same order implies equality
    pairs = pairs[2:] + pairs[:2]
    od2 = OrderedDict(pairs)
    assertNotEqual(od1, od2)  # different order implies inequality
    # comparison to regular dict is not order sensitive
    assertEqual(od1, dict(od2))
    assertEqual(dict(od2), od1)
    # different length implied inequality
    assertNotEqual(od1, OrderedDict(pairs[:-1]))


def test_copying():
    OrderedDict = SyOrderedDict
    # Check that ordered dicts are copyable, deepcopyable, picklable,
    # and have a repr/eval round-trip
    pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
    od = OrderedDict(pairs)

    def check(dup):
        assert dup is not od
        assertEqual(dup, od)
        assertEqual(list(dup.items()), list(od.items()))
        assertEqual(len(dup), len(od))
        assertEqual(type(dup), type(od))

    check(od.copy())
    check(copy.copy(od))
    check(copy.deepcopy(od))

    check(eval(repr(od)))
    update_test = OrderedDict()
    update_test.update(od)
    check(update_test)
    check(OrderedDict(od))


def test_yaml_linkage():
    OrderedDict = SyOrderedDict
    # Verify that __reduce__ is setup in a way that supports PyYAML's dump() feature.
    # In yaml, lists are native but tuples are not.
    pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
    od = OrderedDict(pairs)
    # yaml.dump(od) -->
    # '!!python/object/apply:__main__.OrderedDict\n- - [a, 1]\n  - [b, 2]\n'
    assert all(type(pair) == list for pair in od.__reduce__()[1])


def test_reduce_not_too_fat():
    OrderedDict = SyOrderedDict
    # do not save instance dictionary if not needed
    pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
    od = OrderedDict(pairs)
    assert isinstance(od.__dict__, dict)

    res = od.__reduce__()[2]
    del res["_id"]
    assertEqual(res, {})

    od.x = 10
    assertEqual(od.__dict__["x"], 10)
    assertEqual(od.__reduce__()[2], {"x": 10})


def test_pickle_recursive():
    OrderedDict = SyOrderedDict
    od = OrderedDict()
    od[1] = od

    # pickle directly pulls the module, so we have to fake it
    for proto in range(-1, pickle.HIGHEST_PROTOCOL + 1):
        dup = pickle.loads(pickle.dumps(od, proto))
        assert dup is not od
        assertEqual(list(dup.keys()), [1])
        assert dup[1], dup


def test_repr():
    OrderedDict = SyOrderedDict
    od = OrderedDict([("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)])
    assertEqual(
        repr(od),
        "OrderedDict([('c', 1), ('b', 2), ('a', 3), ('d', 4), ('e', 5), ('f', 6)])",
    )
    assertEqual(eval(repr(od)), od)
    assertEqual(repr(OrderedDict()), "OrderedDict()")


def test_repr_recursive():
    OrderedDict = SyOrderedDict
    # See issue #9826
    od = OrderedDict.FromKeys("abc")
    od["x"] = od
    assertEqual(
        repr(od),
        f"OrderedDict([('a', {repr(SyNone)}), ('b', {repr(SyNone)}), ('c', {repr(SyNone)}), ('x', ...)])",
    )


def test_repr_recursive_values():
    OrderedDict = SyOrderedDict
    od = OrderedDict()
    od[42] = od.values()
    r = repr(od)
    # Cannot perform a stronger test, as the contents of the repr
    # are implementation-dependent.  All we can say is that we
    # want a str result, not an exception of any sort.
    assert isinstance(r, str)
    od[42] = od.items()
    r = repr(od)
    # Again.
    assert isinstance(r, str)


def test_setdefault():
    OrderedDict = SyOrderedDict
    pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
    shuffle(pairs)
    od = OrderedDict(pairs)
    pair_order = list(od.items())
    assertEqual(od.setdefault("a", 10), 3)
    # make sure order didn't change
    assertEqual(list(od.items()), pair_order)
    assertEqual(od.setdefault("x", 10), 10)
    # make sure 'x' is added to the end
    assertEqual(list(od.items())[-1], ("x", 10))
    assertEqual(od.setdefault("g", default=9), 9)

    # make sure setdefault still works when __missing__ is defined
    class Missing(OrderedDict):
        def __missing__(self, key):
            return 0

    assertEqual(Missing().setdefault(5, 9), 9)


def test_reinsert():
    OrderedDict = SyOrderedDict
    # Given insert a, insert b, delete a, re-insert a,
    # verify that a is now later than b.
    od = OrderedDict()
    od["a"] = 1
    od["b"] = 2
    del od["a"]
    assertEqual(list(od.items()), [("b", 2)])
    od["a"] = 1
    assertEqual(list(od.items()), [("b", 2), ("a", 1)])


def test_move_to_end():
    OrderedDict = SyOrderedDict
    od = OrderedDict.FromKeys("abcde")
    assertEqual(list(od), list("abcde"))
    od.move_to_end("c")
    assertEqual(list(od), list("abdec"))
    od.move_to_end("c", 0)
    assertEqual(list(od), list("cabde"))
    od.move_to_end("c", 0)
    assertEqual(list(od), list("cabde"))
    od.move_to_end("e")
    assertEqual(list(od), list("cabde"))
    od.move_to_end("b", last=False)
    assertEqual(list(od), list("bcade"))
    with pytest.raises(KeyError):
        od.move_to_end("x")
    with pytest.raises(KeyError):
        od.move_to_end("x", 0)


def test_move_to_end_issue25406():
    OrderedDict = SyOrderedDict
    od = OrderedDict.FromKeys("abc")
    od.move_to_end("c", last=False)
    assertEqual(list(od), list("cab"))
    od.move_to_end("a", last=False)
    assertEqual(list(od), list("acb"))

    od = OrderedDict.FromKeys("abc")
    od.move_to_end("a")
    assertEqual(list(od), list("bca"))
    od.move_to_end("c")
    assertEqual(list(od), list("bac"))


def test_sizeof():
    OrderedDict = SyOrderedDict
    # Wimpy test: Just verify the reported size is larger than a regular dict
    d = dict(a=1)
    od = OrderedDict(**d)
    assert sys.getsizeof(od) > sys.getsizeof(d)


def test_views():
    OrderedDict = SyOrderedDict
    # See http://bugs.python.org/issue24286
    s = "the quick brown fox jumped over a lazy dog yesterday before dawn".split()
    od = OrderedDict.FromKeys(s)
    assertEqual(list(od.keys()), list(dict(od).keys()))
    assertEqual(list(od.items()), list(dict(od).items()))


def test_override_update():
    OrderedDict = SyOrderedDict
    # Verify that subclasses can override update() without breaking __init__()

    class MyOD(OrderedDict):
        def update(self, *args, **kwds):
            raise Exception()

    items = [("a", 1), ("c", 3), ("b", 2)]
    assertEqual(list(MyOD(items).items()), items)


def test_highly_nested():
    # Issues 25395 and 35983: test that the trashcan mechanism works
    # correctly for OrderedDict: deleting a highly nested OrderDict
    # should not crash Python.
    OrderedDict = SyOrderedDict
    obj = None
    for _ in range(1000):
        obj = OrderedDict([(None, obj)])
    del obj
    gc.collect()


def test_highly_nested_subclass():
    # Issues 25395 and 35983: test that the trashcan mechanism works
    # correctly for OrderedDict: deleting a highly nested OrderDict
    # should not crash Python.
    OrderedDict = SyOrderedDict
    deleted = []

    class MyOD(OrderedDict):
        def __del__(self):
            deleted.append(self.i)

    obj = None
    for i in range(100):
        obj = MyOD([(None, obj)])
        obj.i = i
    del obj
    gc.collect()
    assertEqual(deleted, list(reversed(range(100))))


def test_delitem_hash_collision():
    OrderedDict = SyOrderedDict

    class Key:
        def __init__(self, hash):
            self._hash = hash
            self.value = str(id(self))

        def __hash__(self):
            return self._hash

        def __eq__(self, other):
            try:
                return self.value == other.value
            except AttributeError:
                return False

        def __repr__(self):
            return self.value

    def blocking_hash(hash):
        # See the collision-handling in lookdict (in Objects/dictobject.c).
        MINSIZE = 8
        i = hash & MINSIZE - 1
        return (i << 2) + i + hash + 1

    COLLIDING = 1

    key = Key(COLLIDING)
    colliding = Key(COLLIDING)
    blocking = Key(blocking_hash(COLLIDING))

    od = OrderedDict()
    od[key] = ...
    od[blocking] = ...
    od[colliding] = ...
    od["after"] = ...

    del od[blocking]
    del od[colliding]
    assertEqual(list(od.items()), [(key, ...), ("after", ...)])


def test_issue24347():
    OrderedDict = SyOrderedDict

    class Key:
        def __hash__(self):
            return randrange(100000)

    od = OrderedDict()
    for i in range(100):
        key = Key()
        od[key] = i

    # These should not crash.
    with pytest.raises(RuntimeError):
        list(od.values())
    with pytest.raises(RuntimeError):
        list(od.items())
    with pytest.raises(RuntimeError):
        repr(od)
    with pytest.raises(KeyError):
        od.copy()


def test_issue24348():
    OrderedDict = SyOrderedDict

    class Key:
        def __hash__(self):
            return 1

    od = OrderedDict()
    od[Key()] = 0
    # This should not crash.
    od.popitem()


def test_issue24667():
    """
    dict resizes after a certain number of insertion operations,
    whether or not there were deletions that freed up slots in the
    hash table.  During fast node lookup, OrderedDict must correctly
    respond to all resizes, even if the current "size" is the same
    as the old one.  We verify that here by forcing a dict resize
    on a sparse odict and then perform an operation that should
    trigger an odict resize (e.g. popitem).  One key aspect here is
    that we will keep the size of the odict the same at each popitem
    call.  This verifies that we handled the dict resize properly.
    """
    OrderedDict = SyOrderedDict

    od = OrderedDict()
    for c0 in "0123456789ABCDEF":
        for c1 in "0123456789ABCDEF":
            if len(od) == 4:
                # This should not raise a KeyError.
                od.popitem(last=False)
            key = c0 + c1
            od[key] = key


# Direct use of dict methods


def test_dict_setitem():
    OrderedDict = SyOrderedDict
    od = OrderedDict()
    dict.__setitem__(od, "spam", 1)
    assertNotIn("NULL", repr(od))


def test_dict_delitem():
    OrderedDict = SyOrderedDict
    od = OrderedDict()
    od["spam"] = 1
    od["ham"] = 2
    dict.__delitem__(od, "spam")
    with pytest.raises(RuntimeError):
        repr(od)


def test_dict_clear():
    OrderedDict = SyOrderedDict
    od = OrderedDict()
    od["spam"] = 1
    od["ham"] = 2
    dict.clear(od)
    assertNotIn("NULL", repr(od))


def test_dict_pop():
    OrderedDict = SyOrderedDict
    od = OrderedDict()
    od["spam"] = 1
    od["ham"] = 2
    dict.pop(od, "spam")
    with pytest.raises(RuntimeError):
        repr(od)


def test_dict_popitem():
    OrderedDict = SyOrderedDict
    od = OrderedDict()
    od["spam"] = 1
    od["ham"] = 2
    dict.popitem(od)
    with pytest.raises(RuntimeError):
        repr(od)


def test_dict_setdefault():
    OrderedDict = SyOrderedDict
    od = OrderedDict()
    dict.setdefault(od, "spam", 1)
    assertNotIn("NULL", repr(od))


def test_dict_update():
    OrderedDict = SyOrderedDict
    od = OrderedDict()
    dict.update(od, [("spam", 1)])
    assertNotIn("NULL", repr(od))


def test_reference_loop():
    # Issue 25935
    OrderedDict = SyOrderedDict

    class A:
        od = OrderedDict()

    A.od[A] = None
    r = weakref.ref(A)
    del A
    gc.collect()
    assert r() is None


def test_ordered_dict_items_result_gc():
    # bpo-42536: OrderedDict.items's tuple-reuse speed trick breaks the GC's
    # assumptions about what can be untracked. Make sure we re-track result
    # tuples whenever we reuse them.
    it = iter(SyOrderedDict({None: []}).items())
    gc.collect()
    # That GC collection probably untracked the recycled internal result
    # tuple, which is initialized to (None, None). Make sure it's re-tracked
    # when it's mutated and returned from __next__:
    assert gc.is_tracked(next(it))


def test_key_change_during_iteration():
    OrderedDict = SyOrderedDict

    od = OrderedDict.FromKeys("abcde")
    assertEqual(list(od), list("abcde"))
    with pytest.raises(RuntimeError):
        for i, k in enumerate(od):
            od.move_to_end(k)
            assert i < 5
    with pytest.raises(RuntimeError):
        for k in od:
            od["f"] = None
    with pytest.raises(RuntimeError):
        for k in od:
            del od["c"]
    assertEqual(list(od), list("bdeaf"))


def test_weakref_list_is_not_traversed():
    # Check that the weakref list is not traversed when collecting
    # OrderedDict objects. See bpo-39778 for more information.

    gc.collect()

    x = SyOrderedDict()
    x.cycle = x

    cycle = []
    cycle.append(cycle)

    x_ref = weakref.ref(x)
    cycle.append(x_ref)

    del x, cycle, x_ref

    gc.collect()



# stdlib
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft relative
from ..... import deserialize
from ..... import serialize
from .....logger import traceback_and_raise
from .....proto.core.node.domain.service.request_answer_message_pb2 import (
    RequestAnswerMessage as RequestAnswerMessage_PB,
)
from .....proto.core.node.domain.service.request_answer_response_pb2 import (
    RequestAnswerResponse as RequestAnswerResponse_PB,
)
from ....common import UID
from ....common.message import ImmediateSyftMessageWithReply
from ....common.message import ImmediateSyftMessageWithoutReply
from ....common.serde.serializable import bind_protobuf
from ....io.address import Address
from ...abstract.node import AbstractNode
from ...common.service.node_service import ImmediateNodeServiceWithReply
from .request_message import RequestStatus


@bind_protobuf
class RequestAnswerMessage(ImmediateSyftMessageWithReply):
    __slots__ = ["request_id"]

    def __init__(self, request_id: UID, reply_to: Address, address: Address):
        super().__init__(reply_to, address)
        self.request_id = request_id

    def _object2proto(self) -> RequestAnswerMessage_PB:
        msg = RequestAnswerMessage_PB()
        msg.request_id.CopyFrom(serialize(obj=self.request_id))
        msg.address.CopyFrom(serialize(obj=self.address))
        msg.reply_to.CopyFrom(serialize(obj=self.reply_to))
        return msg

    @staticmethod
    def _proto2object(proto: RequestAnswerMessage_PB) -> "RequestAnswerMessage":
        return RequestAnswerMessage(
            request_id=deserialize(blob=proto.request_id),
            address=deserialize(blob=proto.address),
            reply_to=deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return RequestAnswerMessage_PB


@bind_protobuf
class RequestAnswerResponse(ImmediateSyftMessageWithoutReply):

    __slots__ = ["status", "request_id"]

    def __init__(self, status: RequestStatus, request_id: UID, address: Address):
        super().__init__(address)
        self.status = status
        self.request_id = request_id

    def _object2proto(self) -> RequestAnswerResponse_PB:
        msg = RequestAnswerResponse_PB()
        msg.request_id.CopyFrom(serialize(obj=self.request_id))
        msg.status = self.status.value
        msg.address.CopyFrom(serialize(obj=self.address))
        return msg

    @staticmethod
    def _proto2object(proto: RequestAnswerResponse_PB) -> "RequestAnswerResponse":
        request_response = RequestAnswerResponse(
            status=RequestStatus(proto.status),
            request_id=deserialize(blob=proto.request_id),
            address=deserialize(blob=proto.address),
        )
        return request_response

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return RequestAnswerResponse_PB


class RequestAnswerMessageService(ImmediateNodeServiceWithReply):
    @staticmethod
    def message_handler_types() -> List[type]:
        return [RequestAnswerMessage]

    @staticmethod
    def process(
        node: AbstractNode,
        msg: RequestAnswerMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> RequestAnswerResponse:
        if verify_key is None:
            traceback_and_raise(
                ValueError(
                    "Can't process Request service without a given " "verification key"
                )
            )
        status = node.get_request_status(message_request_id=msg.request_id)  # type: ignore
        address = msg.reply_to
        return RequestAnswerResponse(
            request_id=msg.request_id, address=address, status=status
        )
